import tensorflow as tf
from tensorflow.python import keras
from tensorflow.python.keras import backend as K

custom_objects = tf.keras.utils.get_custom_objects()

L = tf.keras.layers
initializers = keras.initializers


def integerize_shape(func):
    """ decorator, to ensure that input_shape must be int or None
    """

    def convert(item):
        if hasattr(item, '__iter__'):
            return [convert(i) for i in item]
        elif hasattr(item, 'value'):
            return item.value
        else:
            return item

    def new_func(self, input_shape):
        input_shape = convert(input_shape)
        return func(self, input_shape)

    return new_func


def sequence_masking(x, mask, mode=0, axis=None):
    """ is the function of sequence condition mask
    mask: 0-1 matrix in the shape of (batch_size, seq_len);
    mode: If it is 0, it is directly multiplied by the mask;
          If it is 1, subtract a large positive number from the padding part.
    axis: the axis of the sequence, the default is 1;
    """
    if mask is None or mode not in [0, 1]:
        return x
    else:
        if axis is None:
            axis = 1
        if axis == -1:
            axis = K.ndim(x) - 1
        assert axis > 0, 'axis muse be greater than 0'
        for _ in range(axis - 1):
            mask = K.expand_dims(mask, 1)
        for _ in range(K.ndim(x) - K.ndim(mask) - axis + 1):
            mask = K.expand_dims(mask, K.ndim(mask))
        if mode == 0:
            return x * mask
        else:
            return x - (1 - mask) * 1e12


class ConditionalRandomField(L.Layer):
    """Pure Keras realizes CRF layer
       The CRF layer is essentially a loss calculation layer with training parameters.
    """

    def __init__(self, lr_multiplier=1, **kwargs):
        super(ConditionalRandomField, self).__init__(**kwargs)
        self.lr_multiplier = lr_multiplier  # The magnification of the current layer learning rate

    @integerize_shape
    def build(self, input_shape):
        super(ConditionalRandomField, self).build(input_shape)
        output_dim = input_shape[-1]
        self._trans = self.add_weight(
            name='trans',
            shape=(output_dim, output_dim),
            initializer='glorot_uniform',
            trainable=True
        )
        if self.lr_multiplier != 1:
            K.set_value(self._trans, K.eval(self._trans) / self.lr_multiplier)

    @property
    def trans(self):
        if self.lr_multiplier != 1:
            return self.lr_multiplier * self._trans
        else:
            return self._trans

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, inputs, mask=None):
        if mask is not None:
            mask = K.cast(mask, K.floatx())

        return sequence_masking(inputs, mask, 1, 1)

    def target_score(self, y_true, y_pred):
        """Calculate the relative probability of the target path (not normalized yet)
        Key points: score by label plus transition probability score.
        """
        point_score = tf.einsum('bni,bni->b', y_true, y_pred)  # 逐标签得分
        trans_score = tf.einsum(
            'bni,ij,bnj->b', y_true[:, :-1], self.trans, y_true[:, 1:]
        )  # Tag transfer score
        return point_score + trans_score

    def log_norm_step(self, inputs, states):
        """Recursively calculate the normalization factor
        Key points: 1. Recursive calculation; 2. Use logsumexp to avoid overflow.
        """
        inputs, mask = inputs[:, :-1], inputs[:, -1:]
        states = K.expand_dims(states[0], 2)  # (batch_size, output_dim, 1)
        trans = K.expand_dims(self.trans, 0)  # (1, output_dim, output_dim)
        outputs = tf.reduce_logsumexp(
            states + trans, 1
        )  # (batch_size, output_dim)
        outputs = outputs + inputs
        outputs = mask * outputs + (1 - mask) * states[:, :, 0]
        return outputs, [outputs]

    def dense_loss(self, y_true, y_pred):
        """y_true needs to be in one hot form
        """
        # Export mask and convert texts type
        mask = K.all(K.greater(y_pred, -1e6), axis=2, keepdims=True)
        mask = K.cast(mask, K.floatx())
        # Calculate the target score
        y_true, y_pred = y_true * mask, y_pred * mask
        target_score = self.target_score(y_true, y_pred)
        # Recursively calculate log Z
        init_states = [y_pred[:, 0]]
        y_pred = K.concatenate([y_pred, mask], axis=2)
        input_length = K.int_shape(y_pred[:, 1:])[1]
        log_norm, _, _ = K.rnn(
            self.log_norm_step,
            y_pred[:, 1:],
            init_states,
            input_length=input_length
        )  # Log Z vector of the last step
        log_norm = tf.reduce_logsumexp(log_norm, 1)
        # Calculate loss -log p
        return log_norm - target_score

    def sparse_loss(self, y_true, y_pred):
        """y_true needs to be an integer (not one hot)
        """
        # y_true Need to re-clarify shape and dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # Convert to one hot
        y_true = K.one_hot(y_true, K.shape(self.trans)[0])
        return self.dense_loss(y_true, y_pred)

    def dense_accuracy(self, y_true, y_pred):
        """Displays the frame-by-frame accuracy function during training, excluding the influence of mask
        Here y_true needs to be in one hot form
        """
        y_true = K.argmax(y_true, 2)
        return self.sparse_accuracy(y_true, y_pred)

    def sparse_accuracy(self, y_true, y_pred):
        """Displays the frame-by-frame accuracy function during training, excluding the influence of mask
        Here y_true needs to be in integer form (not one hot)
        """
        # Export mask and convert texts type
        mask = K.all(K.greater(y_pred, -1e6), axis=2)
        mask = K.cast(mask, K.floatx())
        # y_true Need to re-clarify shape and dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # Take the largest label by label to roughly evaluate the training effect
        y_pred = K.cast(K.argmax(y_pred, 2), 'int32')
        isequal = K.cast(K.equal(y_true, y_pred), K.floatx())
        return K.sum(isequal * mask) / K.sum(mask)

    def get_config(self):
        config = {
            'lr_multiplier': self.lr_multiplier,
        }
        base_config = super(ConditionalRandomField, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class MaximumEntropyMarkovModel(L.Layer):
    """ (Bidirectional) Maximum Entropy Hidden Markov Model
    Function and usage are similar to CRF, but faster and simpler than CRF.
    """

    def __init__(self, lr_multiplier=1, hidden_dim=None, **kwargs):
        super(MaximumEntropyMarkovModel, self).__init__(**kwargs)
        self.lr_multiplier = lr_multiplier  # The magnification of the current layer learning rate
        self.hidden_dim = hidden_dim  # If it is not None, decompose the transfer matrix with low rank

    @integerize_shape
    def build(self, input_shape):
        super(MaximumEntropyMarkovModel, self).build(input_shape)
        output_dim = input_shape[-1]

        if self.hidden_dim is None:
            self._trans = self.add_weight(
                name='trans',
                shape=(output_dim, output_dim),
                initializer='glorot_uniform',
                trainable=True
            )
            if self.lr_multiplier != 1:
                K.set_value(
                    self._trans,
                    K.eval(self._trans) / self.lr_multiplier
                )
        else:
            self._l_trans = self.add_weight(
                name='l_trans',
                shape=(output_dim, self.hidden_dim),
                initializer='glorot_uniform',
                trainable=True
            )
            self._r_trans = self.add_weight(
                name='r_trans',
                shape=(output_dim, self.hidden_dim),
                initializer='glorot_uniform',
                trainable=True
            )

            if self.lr_multiplier != 1:
                K.set_value(
                    self._l_trans,
                    K.eval(self._l_trans) / self.lr_multiplier
                )
                K.set_value(
                    self._r_trans,
                    K.eval(self._r_trans) / self.lr_multiplier
                )

    @property
    def trans(self):
        if self.lr_multiplier != 1:
            return self.lr_multiplier * self._trans
        else:
            return self._trans

    @property
    def l_trans(self):
        if self.lr_multiplier != 1:
            return self.lr_multiplier * self._l_trans
        else:
            return self._l_trans

    @property
    def r_trans(self):
        if self.lr_multiplier != 1:
            return self.lr_multiplier * self._r_trans
        else:
            return self._r_trans

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, inputs, mask=None):
        if mask is not None:
            mask = K.cast(mask, K.floatx())

        return sequence_masking(inputs, mask, 1, 1)

    def reverse_sequence(self, inputs, mask=None):
        if mask is None:
            return [x[:, ::-1] for x in inputs]
        else:
            length = K.cast(K.sum(mask, 1), 'int32')
            return [tf.reverse_sequence(x, length, seq_axis=1) for x in inputs]

    def basic_loss(self, y_true, y_pred, go_backwards=False):
        """y_true needs to be an integer (not one hot)
        """
        # Export mask and convert texts type
        mask = K.all(K.greater(y_pred, -1e6), axis=2)
        mask = K.cast(mask, K.floatx())
        # y_true Need to re-clarify shape and dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # Reverse related
        if self.hidden_dim is None:
            if go_backwards:  # Whether to reverse the sequence
                y_true, y_pred = self.reverse_sequence([y_true, y_pred], mask)
                trans = K.transpose(self.trans)
            else:
                trans = self.trans
            histoty = K.gather(trans, y_true)
        else:
            if go_backwards:  # Whether to reverse the sequence
                y_true, y_pred = self.reverse_sequence([y_true, y_pred], mask)
                r_trans, l_trans = self.l_trans, self.r_trans
            else:
                l_trans, r_trans = self.l_trans, self.r_trans
            histoty = K.gather(l_trans, y_true)
            histoty = tf.einsum('bnd,kd->bnk', histoty, r_trans)
        # Calculate loss
        histoty = K.concatenate([y_pred[:, :1], histoty[:, :-1]], 1)
        y_pred = (y_pred + histoty) / 2
        loss = K.sparse_categorical_crossentropy(
            y_true, y_pred, from_logits=True
        )
        return K.sum(loss * mask) / K.sum(mask)

    def sparse_loss(self, y_true, y_pred):
        """y_true needs to be an integer (not one hot)
        """
        loss = self.basic_loss(y_true, y_pred, False)
        loss = loss + self.basic_loss(y_true, y_pred, True)
        return loss / 2

    def dense_loss(self, y_true, y_pred):
        """y_true needs to be in one hot form
        """
        y_true = K.argmax(y_true, 2)
        return self.sparse_loss(y_true, y_pred)

    def basic_accuracy(self, y_true, y_pred, go_backwards=False):
        """Displays the frame-by-frame accuracy function during training, excluding the influence of mask
        Here y_true needs to be in integer form (not one hot)
        """
        # Export mask and convert texts type
        mask = K.all(K.greater(y_pred, -1e6), axis=2)
        mask = K.cast(mask, K.floatx())
        # y_true Need to re-clarify shape and dtype
        y_true = K.reshape(y_true, K.shape(y_pred)[:-1])
        y_true = K.cast(y_true, 'int32')
        # Reverse related
        if self.hidden_dim is None:
            if go_backwards:  # Whether to reverse the sequence
                y_true, y_pred = self.reverse_sequence([y_true, y_pred], mask)
                trans = K.transpose(self.trans)
            else:
                trans = self.trans
            histoty = K.gather(trans, y_true)
        else:
            if go_backwards:  # Whether to reverse the sequence
                y_true, y_pred = self.reverse_sequence([y_true, y_pred], mask)
                r_trans, l_trans = self.l_trans, self.r_trans
            else:
                l_trans, r_trans = self.l_trans, self.r_trans
            histoty = K.gather(l_trans, y_true)
            histoty = tf.einsum('bnd,kd->bnk', histoty, r_trans)
        # Calculate the label-by-label accuracy
        histoty = K.concatenate([y_pred[:, :1], histoty[:, :-1]], 1)
        y_pred = (y_pred + histoty) / 2
        y_pred = K.cast(K.argmax(y_pred, 2), 'int32')
        isequal = K.cast(K.equal(y_true, y_pred), K.floatx())
        return K.sum(isequal * mask) / K.sum(mask)

    def sparse_accuracy(self, y_true, y_pred):
        """Displays the frame-by-frame accuracy function during training, excluding the influence of mask
        Here y_true needs to be in integer form (not one hot)
        """
        accuracy = self.basic_accuracy(y_true, y_pred, False)
        accuracy = accuracy + self.basic_accuracy(y_true, y_pred, True)
        return accuracy / 2

    def dense_accuracy(self, y_true, y_pred):
        """Displays the frame-by-frame accuracy function during training, excluding the influence of mask
        Here y_true needs to be in one hot form
        """
        y_true = K.argmax(y_true, 2)
        return self.sparse_accuracy(y_true, y_pred)

    def get_config(self):
        config = {
            'lr_multiplier': self.lr_multiplier,
            'hidden_dim': self.hidden_dim,
        }
        base_config = super(MaximumEntropyMarkovModel, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



"""
A Vocabulary maps strings to integers, allowing for strings to be mapped to an
out-of-vocabulary token.
"""

import codecs
import copy
import logging
import os
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Union

import tqdm

from kolibri.config import ModelConfig

logger = logging.getLogger(__name__)

DEFAULT_NON_PADDED_NAMESPACES = ("*tags", "*y_values")
DEFAULT_PADDING_TOKEN = "__PADDING__"
DEFAULT_OOV_TOKEN = "__UNKNOWN__"
NAMESPACE_PADDING_FILE = "non_padded_namespaces.txt"


def pop_max_vocab_size(params: ModelConfig):
    """
    max_vocab_size: maximum size of the vocabulary, not including the __UNKNOWN__ token..
    """
    size = params.get("max_vocab_size", None)

    if isinstance(size, dict):
        # This is the Dict[str, int] case.
        return size
    elif size is not None:
        # This is the int / str case.
        return int(size)
    else:
        return None


class Vocabulary():
    """
    Maps strings to integers and integrers to strings, allowing for strings to be mapped to an
    out-of-vocabulary token..
    """

    default_implementation = "default"

    def __init__(
            self,
            counter: Dict[str, Dict[str, int]] = None,
            min_count: Dict[str, int] = None,
            max_vocab_size: Union[int, Dict[str, int]] = None,
            non_padded_namespaces: Iterable[str] = DEFAULT_NON_PADDED_NAMESPACES,
            tokens_to_add: Dict[str, List[str]] = None,
            padding_token: Optional[str] = DEFAULT_PADDING_TOKEN,
            oov_token: Optional[str] = DEFAULT_OOV_TOKEN,
            recemmended_padding: int = None
    ) -> None:

        self._padding_token = padding_token if padding_token is not None else DEFAULT_PADDING_TOKEN
        self._oov_token = oov_token if oov_token is not None else DEFAULT_OOV_TOKEN

        self._non_padded_namespaces = set(non_padded_namespaces)

        self._token_to_index = defaultdict()
        self._index_to_token = defaultdict()

        self.recommended_padding_lengths = recemmended_padding
        self._retained_counter: Optional[Dict[str, Dict[str, int]]] = None
        # Made an empty vocabulary, now extend it.
        self._extend(
            counter,
            min_count,
            max_vocab_size,
            non_padded_namespaces,
            tokens_to_add,
        )

    def __getstate__(self):
        """
        Need to sanitize defaultdict and defaultdict-like objects
        by converting them to vanilla dicts when we pickle the vocabulary.
        """
        state = copy.copy(self.__dict__)
        state["_token_to_index"] = dict(state["_token_to_index"])
        state["_index_to_token"] = dict(state["_index_to_token"])

        if "_retained_counter" in state:
            state["_retained_counter"] = {
                key: dict(value) for key, value in state["_retained_counter"].items()
            }

        return state

    def __setstate__(self, state):
        """
        Conversely, when we unpickle, we need to reload the plain dicts
        into our special DefaultDict subclasses.
        """

        self.__dict__ = copy.copy(state)
        self._token_to_index = defaultdict()
        self._token_to_index.update(state["_token_to_index"])
        self._index_to_token = defaultdict()
        self._index_to_token.update(state["_index_to_token"])

    def save_to_files(self, directory: str) -> None:
        """
        Persist this Vocabulary to files so it can be reloaded later.
        Each namespace corresponds to one file.

        Parameters
        ----------
        directory : ``str``
            The directory where we save the serialized vocabulary.
        """
        os.makedirs(directory, exist_ok=True)
        if os.listdir(directory):
            logging.warning("vocabulary serialization directory %s is not empty", directory)

        with codecs.open(
                os.path.join(directory, NAMESPACE_PADDING_FILE), "w", "utf-8"
        ) as namespace_file:
            for namespace_str in self._non_padded_namespaces:
                print(namespace_str, file=namespace_file)

        for namespace, mapping in self._index_to_token.items():
            # Each namespace gets written to its own file, in index order.
            with codecs.open(
                    os.path.join(directory, namespace + ".txt"), "w", "utf-8"
            ) as token_file:
                num_tokens = len(mapping)
                start_index = 1 if mapping[0] == self._padding_token else 0
                for i in range(start_index, num_tokens):
                    print(mapping[i].replace("\n", "__NEWLINE__"), file=token_file)

    @classmethod
    def from_files(
            cls,
            directory: str,
            padding_token: Optional[str] = DEFAULT_PADDING_TOKEN,
            oov_token: Optional[str] = DEFAULT_OOV_TOKEN,
    ) -> "Vocabulary":
        """
        Loads a ``Vocabulary`` that was serialized using ``save_to_files``.

        Parameters
        ----------
        directory : ``str``
            The directory containing the serialized vocabulary.
        """
        logger.info("Loading token dictionary from %s.", directory)
        padding_token = padding_token if padding_token is not None else DEFAULT_PADDING_TOKEN
        oov_token = oov_token if oov_token is not None else DEFAULT_OOV_TOKEN
        with codecs.open(
                os.path.join(directory, NAMESPACE_PADDING_FILE), "r", "utf-8"
        ) as namespace_file:
            non_padded_namespaces = [namespace_str.strip() for namespace_str in namespace_file]

        vocab = cls(
            non_padded_namespaces=non_padded_namespaces,
            padding_token=padding_token,
            oov_token=oov_token,
        )

        # Check every file in the directory.
        for namespace_filename in os.listdir(directory):
            if namespace_filename == NAMESPACE_PADDING_FILE:
                continue
            if namespace_filename.startswith("."):
                continue
            namespace = namespace_filename.replace(".txt", "")
            filename = os.path.join(directory, namespace_filename)
            vocab.set_from_file(filename, True, namespace=namespace, oov_token=oov_token)

        return vocab

    def set_from_file(
            self,
            filename: str,
            is_padded: bool = True,
            oov_token: str = DEFAULT_OOV_TOKEN,
            namespace: str = "tokens",
    ):
        """
        If you already have a vocabulary file for a trained model somewhere, and you really want to
        use that vocabulary file instead of just setting the vocabulary from a dataset, for
        whatever reason, you can do that with this method.  You must specify the namespace to use,
        and we assume that you want to use padding and OOV tokens for this.

        Parameters
        ----------
        filename : ``str``
            The file containing the vocabulary to load.  It should be formatted as one token per
            line, with nothing else in the line.  The index we assign to the token is the line
            number in the file (1-indexed if ``is_padded``, 0-indexed otherwise).  Note that this
            file should contain the OOV token string!
        is_padded : ``bool``, optional (default=True)
            Is this vocabulary padded?  For token / word / character vocabularies, this should be
            ``True``; while for tag or label vocabularies, this should typically be ``False``.  If
            ``True``, we add a padding token with index 0, and we enforce that the ``oov_token`` is
            present in the file.
        oov_token : ``str``, optional (default=DEFAULT_OOV_TOKEN)
            What token does this vocabulary use to represent out-of-vocabulary characters?  This
            must show up as a line in the vocabulary file.  When we find it, we replace
            ``oov_token`` with ``self._oov_token``, because we only use one OOV token across
            namespaces.
        namespace : ``str``, optional (default="tokens")
            What namespace should we overwrite with this vocab file?
        """
        if is_padded:
            self._token_to_index[namespace] = {self._padding_token: 0}
            self._index_to_token[namespace] = {0: self._padding_token}
        else:
            self._token_to_index[namespace] = {}
            self._index_to_token[namespace] = {}
        with codecs.open(filename, "r", "utf-8") as input_file:
            lines = input_file.read().split("\n")
            # Be flexible about having final newline or not
            if lines and lines[-1] == "":
                lines = lines[:-1]
            for i, line in enumerate(lines):
                index = i + 1 if is_padded else i
                token = line.replace("@@NEWLINE@@", "\n")
                if token == oov_token:
                    token = self._oov_token
                self._token_to_index[namespace][token] = index
                self._index_to_token[namespace][index] = token
        if is_padded:
            assert self._oov_token in self._token_to_index[namespace], "OOV token not found!"

    @classmethod
    def from_instances(
            cls,
            instances: Iterable["adi.Instance"],
            min_count: Dict[str, int] = None,
            max_vocab_size: Union[int, Dict[str, int]] = None,
            non_padded_namespaces: Iterable[str] = DEFAULT_NON_PADDED_NAMESPACES,
            tokens_to_add: Dict[str, List[str]] = None,
            padding_token: Optional[str] = DEFAULT_PADDING_TOKEN,
            oov_token: Optional[str] = DEFAULT_OOV_TOKEN,
    ) -> "Vocabulary":
        """
        Constructs a vocabulary given a collection of `Instances` and some parameters.
        We count all of the vocabulary items in the instances, then pass those counts
        and the other parameters, to :func:`__init__`.  See that method for a description
        of what the other parameters do.
        """
        logger.info("Fitting token dictionary from dataset.")
        padding_token = padding_token if padding_token is not None else DEFAULT_PADDING_TOKEN
        oov_token = oov_token if oov_token is not None else DEFAULT_OOV_TOKEN
        namespace_token_counts: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
        all_instance_lengths: List[Dict[str, Dict[str, int]]] = []
        padding_lengths: Dict[str, Dict[str, int]] = defaultdict(dict)
        for instance in tqdm.tqdm(instances):
            instance.count_vocab_items(namespace_token_counts)
            all_instance_lengths.append(instance.get_padding_lengths())

        if not all_instance_lengths:
            _recommended_padding = {**padding_lengths}
        else:
            all_field_lengths: Dict[str, List[Dict[str, int]]] = defaultdict(list)
            for instance_lengths in all_instance_lengths:
                for field_name, instance_field_lengths in instance_lengths.items():
                    all_field_lengths[field_name].append(instance_field_lengths)

            for field_name, field_lengths in all_field_lengths.items():
                padding_lengths[field_name] = sorted(field_lengths)[int(0.90 * len(field_lengths))]
            _recommended_padding = {**padding_lengths}
        return cls(
            counter=namespace_token_counts,
            min_count=min_count,
            max_vocab_size=max_vocab_size,
            non_padded_namespaces=non_padded_namespaces,
            tokens_to_add=tokens_to_add,
            padding_token=padding_token,
            oov_token=oov_token,
            recemmended_padding=_recommended_padding
        )

    def _extend(
            self,
            counter: Dict[str, Dict[str, int]] = None,
            min_count: Dict[str, int] = None,
            max_vocab_size: Union[int, Dict[str, int]] = None,
            non_padded_namespaces: Iterable[str] = DEFAULT_NON_PADDED_NAMESPACES,
            pretrained_files: Optional[Dict[str, str]] = None,
            only_include_pretrained_words: bool = False,
            tokens_to_add: Dict[str, List[str]] = None,
            min_pretrained_embeddings: Dict[str, int] = None,
    ) -> None:
        """
        This method is used for extend an already generated vocabulary.
        """
        if not isinstance(max_vocab_size, dict):
            int_max_vocab_size = max_vocab_size
            max_vocab_size = defaultdict(lambda: int_max_vocab_size)  # type: ignore
        min_count = min_count or {}
        pretrained_files = pretrained_files or {}
        min_pretrained_embeddings = min_pretrained_embeddings or {}
        non_padded_namespaces = set(non_padded_namespaces)
        counter = counter or {}
        tokens_to_add = tokens_to_add or {}

        self._retained_counter = counter
        # Make sure vocabulary extension is safe.
        current_namespaces = {*self._token_to_index}
        extension_namespaces = {*counter, *tokens_to_add}

        for namespace in current_namespaces & extension_namespaces:
            # if new namespace was already present
            # Either both should be padded or none should be.
            original_padded = not any(
                namespace_match(pattern, namespace) for pattern in self._non_padded_namespaces
            )
            extension_padded = not any(
                namespace_match(pattern, namespace) for pattern in non_padded_namespaces
            )
            if original_padded != extension_padded:
                raise Exception(
                    "Common namespace {} has conflicting ".format(namespace)
                    + "setting of padded = True/False. "
                    + "Hence extension cannot be done."
                )

        # Add new non-padded namespaces for extension
        self._token_to_index.add_non_padded_namespaces(non_padded_namespaces)
        self._index_to_token.add_non_padded_namespaces(non_padded_namespaces)
        self._non_padded_namespaces.update(non_padded_namespaces)

        for namespace in counter:
            pretrained_set = None
            token_counts = list(counter[namespace].items())
            token_counts.sort(key=lambda x: x[1], reverse=True)
            try:
                max_vocab = max_vocab_size[namespace]
            except KeyError:
                max_vocab = None
            if max_vocab:
                token_counts = token_counts[:max_vocab]
            for token, count in token_counts:
                if pretrained_set is not None:
                    if only_include_pretrained_words:
                        if token in pretrained_set and count >= min_count.get(namespace, 1):
                            self.add_token_to_namespace(token, namespace)
                    elif token in pretrained_set or count >= min_count.get(namespace, 1):
                        self.add_token_to_namespace(token, namespace)
                elif count >= min_count.get(namespace, 1):
                    self.add_token_to_namespace(token, namespace)

        for namespace, tokens in tokens_to_add.items():
            for token in tokens:
                self.add_token_to_namespace(token, namespace)

    def extend_from_instances(
            self, params: ModelConfig, instances: Iterable["adi.Instance"] = ()
    ) -> None:
        """
        Extends an already generated vocabulary using a collection of instances.
        """
        min_count = params.get("min_count", None)
        max_vocab_size = pop_max_vocab_size(params)
        non_padded_namespaces = params.get("non_padded_namespaces", DEFAULT_NON_PADDED_NAMESPACES)
        pretrained_files = params.get("pretrained_files", {})
        min_pretrained_embeddings = params.get("min_pretrained_embeddings", None)
        only_include_pretrained_words = params.get("only_include_pretrained_words", False)
        tokens_to_add = params.get("tokens_to_add", None)

        logger.info("Fitting token dictionary from dataset.")
        namespace_token_counts: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
        for instance in tqdm.tqdm(instances):
            instance.count_vocab_items(namespace_token_counts)
        self._extend(
            counter=namespace_token_counts,
            min_count=min_count,
            max_vocab_size=max_vocab_size,
            non_padded_namespaces=non_padded_namespaces,
            pretrained_files=pretrained_files,
            only_include_pretrained_words=only_include_pretrained_words,
            tokens_to_add=tokens_to_add,
            min_pretrained_embeddings=min_pretrained_embeddings,
        )

    def is_padded(self, namespace: str) -> bool:
        """
        Returns whether or not there are padding and OOV tokens added to the given namespace.
        """
        return self._index_to_token[namespace][0] == self._padding_token

    def add_token_to_namespace(self, token: str, namespace: str = "tokens") -> int:
        """
        Adds ``token`` to the index, if it is not already present.  Either way, we return the index of
        the token.
        """
        if not isinstance(token, str):
            raise ValueError(
                "Vocabulary tokens must be strings, or saving and loading will break."
                "  Got %s (with type %s)" % (repr(token), type(token))
            )
        if token not in self._token_to_index[namespace]:
            index = len(self._token_to_index[namespace])
            self._token_to_index[namespace][token] = index
            self._index_to_token[namespace][index] = token
            return index
        else:
            return self._token_to_index[namespace][token]

    def add_tokens_to_namespace(self, tokens: List[str], namespace: str = "tokens") -> List[int]:
        """
        Adds ``tokens`` to the index, if they are not already present.
        """
        return [self.add_token_to_namespace(token, namespace) for token in tokens]

    def get_index_to_token_vocabulary(self, namespace: str = "tokens") -> Dict[int, str]:
        return self._index_to_token[namespace]

    def get_token_to_index_vocabulary(self, namespace: str = "tokens") -> Dict[str, int]:
        return self._token_to_index[namespace]

    def get_token_index(self, token: str, namespace: str = "tokens") -> int:
        if token in self._token_to_index[namespace]:
            return self._token_to_index[namespace][token]
        else:
            try:
                return self._token_to_index[namespace][self._oov_token]
            except KeyError:
                logger.error("Namespace: %s", namespace)
                logger.error("Token: %s", token)
                raise

    def get_token_from_index(self, index: int, namespace: str = "tokens") -> str:
        return self._index_to_token[namespace][index]

    def get_vocab_size(self, namespace: str = "tokens") -> int:
        return len(self._token_to_index[namespace])

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return False

    def print_statistics(self) -> None:
        if self._retained_counter:
            logger.info(
                "Printed vocabulary statistics are only for the part of the vocabulary generated from instances. "
            )
            print("\n\n----Vocabulary Statistics----\n")
            # Since we don't saved counter info, it is impossible to consider pre-saved portion.
            for namespace in self._retained_counter:
                tokens_with_counts = list(self._retained_counter[namespace].items())
                tokens_with_counts.sort(key=lambda x: x[1], reverse=True)
                print(f"\nTop 10 most frequent tokens in namespace '{namespace}':")
                for token, freq in tokens_with_counts[:10]:
                    print(f"\tToken: {token}\t\tFrequency: {freq}")
                # Now sort by token length_train, not frequency
                tokens_with_counts.sort(key=lambda x: len(x[0]), reverse=True)

                print(f"\nTop 10 longest tokens in namespace '{namespace}':")
                for token, freq in tokens_with_counts[:10]:
                    print(f"\tToken: {token}\t\tlength: {len(token)}\tFrequency: {freq}")

                print(f"\nTop 10 shortest tokens in namespace '{namespace}':")
                for token, freq in reversed(tokens_with_counts[-10:]):
                    print(f"\tToken: {token}\t\tlength: {len(token)}\tFrequency: {freq}")
        else:
            # _retained_counter would be set only if instances were used for vocabulary construction.
            logger.info(
                "Vocabulary statistics cannot be printed since dataset instances were not used for its construction."
            )



from tensorflow.keras import Model
from tensorflow.keras.layers import Input, Dense

from kolibri.dnn.autoencoder.base_autoencoder import BaseAutoencoder


class Autoencoder(BaseAutoencoder):

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'layer_multiplyer': 2,
            'hidden': {
                'units': 40,
                'activation': 'relu'
            },
            'latent':{
                'units': 10,
                'activation': 'relu'
                       },
            'output': {
                'activation': 'relu'
                }
                }


    def build_model_arc(self):

        input_layer = Input(shape=(self.input_dim,))

        encoder = Dense(**self.hyper_parameters['hidden'])(input_layer)
        latent = Dense(**self.hyper_parameters['latent'])(encoder)

        latent_inputs = Input(shape=(self.hyper_parameters['latent']['units'],), name='decoder_input')

        decoder = Dense(**self.hyper_parameters['hidden'])(latent_inputs)
        output = Dense(self.input_dim, **self.hyper_parameters['output'])(decoder)

        self.encoder=Model(inputs=input_layer, outputs=latent)

        self.encoder.summary()
        # instantiate decoder model

        self.decoder = Model(latent_inputs, output, name='decoder')
        self.decoder.summary()

        # autoencoder = encoder + decoder
        # instantiate autoencoder model
        self.autoencoder = Model(input_layer,
                            self.decoder(self.encoder(input_layer)),
                            name='autoencoder')


if __name__ == "__main__":


    from kolibri.data.corpus.corpus import CreditCardFraud



    corpus = CreditCardFraud()
    train_x=corpus.X

    model = Autoencoder()
    model.build_model(train_x[:100])



    epochs = 4
    history = None
    model.fit(train_x[0:20000], epochs=epochs)

    model.evaluate(train_x[0:10000])




import logging
import os
from typing import List
from typing import Text

import joblib
from gensim.models import Phrases

from kolibri.kolibri_component import Component
from kolibri.utils.file import save_to_disk, load_from_disk

logger = logging.getLogger(__name__)

BIGRAM_FILENAME = "kolibri-bigram.model"


class BigramAnalyzer(Component):

    _estimator_type = 'estimator'

    name = 'bigram_analyzer'

    provides = ["ngram"]

    requires = ["text_features"]

    defaults = {
        # We try to find a good number of cross folds to use during
        # class training, this specifies the max number of folds
        "min-count": 5,

        "threshold": 10,

        "retrain-bigram": False
    }

    def __init__(self, component_config=None):
        """Construct a new class classifier using the sklearn framework."""
        super().__init__(component_config=component_config)
        self.bigram_model = None

    @classmethod
    def required_packages(cls):
        # type: () -> List[Text]
        return ["gensim"]

    def train(self, training_data, **kwargs):
        retrain = self.component_config['retrain-bigram']
        if retrain:
            self.train_bigram(training_data)
            self.save_bigram_model()
        elif self.bigram_model:
            [self.process(d) for d in training_data]
        else:
            self.bigram_model = self.load_bigram_model()
            [self.process(d) for d in training_data]

    def fit(self, X, y):
        retrain = self.component_config['retrain-bigram']
        if retrain:
            self.train_bigram(X)
            self.save_bigram_model()
        else:
            self.bigram_model = self.load_bigram_model()
        return self

    def transform(self, data, **kwargs):
        return [self.bigram_model[d] for d in data]

    def process(self, document, **kwargs):
        return self.bigram_model[document.sentences]

    def train_bigram(self, trainingdata):
        """
        Train the word2vec model on a directory with text files.
        :param train_dir: directory with '.txt' files
        :param vec_dim: dimensionality of the word vectors

        :return: trained gensim model
        """

        class SentenceIterator(object):
            def __init__(self, data):
                self.training_data = data

            def __iter__(self):
                for document in self.training_data:
                    for sentence in document.sentences:
                        yield sentence

        self.bigram_model = Phrases(SentenceIterator(trainingdata), min_count=self.component_config['min-count'],
                                    threshold=self.component_config['threshold'])

        return self.bigram_model

    def save_bigram_model(self, overwrite=True):
        """ Save the word2vec model to a file """
        if not self.bigram_model:
            raise ValueError("Can't save the word2vec model, " + \
                             "it has not been trained yet")
        save_to_disk(os.path.join(self.component_config["output-folder"], BIGRAM_FILENAME), self.bigram_model,
                     overwrite=overwrite)

    def load_bigram_model(self):
        """ Load the word2vec model from a file """
        self.bigram_model = load_from_disk(os.path.join(self.component_config["output-folder"], BIGRAM_FILENAME))
        return self.bigram_model

    @classmethod
    def load(cls, model_dir=None, model_metadata=None, cached_component=None, **kwargs):

        meta = model_metadata.for_component(cls.name)
        file_name = meta.get("bigram_file", BIGRAM_FILENAME)
        classifier_file = os.path.join(model_dir, file_name)

        if os.path.exists(classifier_file):
            model = joblib.load(classifier_file)

            return model
        else:
            return cls(meta)

    def persist(self, model_dir):
        """Persist this model into the passed directory."""

        bigram_file = os.path.join(model_dir, BIGRAM_FILENAME)
        joblib.dump(self, bigram_file)

        return {
            "bigram_file": BIGRAM_FILENAME
        }

    def get_info(self):
        return "brigram analyzer"



import json
import logging
import os

from tensorflow.keras.utils import get_file

from kolibri import settings as k

LOGGER = logging.getLogger(__name__)

STOP_WORDS_CACHE = {}


class Stopwords():
    PACKAGE = 'stop-words'
    URL_MODEL = "https://www.dropbox.com/s/zoyhw8wpzvj78a4/stop-words.tar.gz?dl=1"

    def __init__(self):
        """
        :param with_additional_file: Allows to load LEFFF without the additional file. (Default: True)
        :type with_additional_file: bool
        :param load_only_pos: Allows to load LEFFF with only some pos tags: WordNet pos tags [a, r, n, v]. (Default: all)
        :type load_only_pos: list
        """
        #        data_file_path = os.path.dirname(os.path.realpath(__file__))

        self.stopwords_path = get_file(self.PACKAGE,
                                       self.URL_MODEL,
                                       cache_dir=k.DATA_PATH,
                                       cache_subdir='corpora',
                                       untar=True)
        with open(os.path.join(self.stopwords_path, 'languages.json'), 'rb') as map_file:
            buffer = map_file.read()
            buffer = buffer.decode('ascii')
            self.LANGUAGE_MAPPING = json.loads(buffer)

        self.AVAILABLE_LANGUAGES = list(self.LANGUAGE_MAPPING.values())


class StopWordError(Exception):
    pass


sw = Stopwords()


def get_stop_words(language, cache=True, aggressive=False):
    """
    :type language: basestring
    :rtype: list
    """
    try:
        language = sw.LANGUAGE_MAPPING[language]
    except KeyError:
        if language not in sw.AVAILABLE_LANGUAGES:
            raise StopWordError('{0}" language is unavailable.'.format(
                language
            ))

    if cache and language in STOP_WORDS_CACHE:
        return STOP_WORDS_CACHE[language]
    language_name = language
    if aggressive:
        language_name + "-aggressive"
    language_filename = os.path.join(sw.stopwords_path, language_name + '.txt')
    try:
        with open(language_filename, 'rb') as language_file:
            stop_words = [line.decode('utf-8').strip()
                          for line in language_file.readlines()]
    except IOError:
        raise StopWordError(
            '{0}" file is unreadable, check your installation.'.format(
                language_filename
            )
        )

    if cache:
        STOP_WORDS_CACHE[language] = stop_words

    return stop_words


if __name__ == "__main__":
    stp = get_stop_words('en')

    print(stp)



from kolibri.tasks.classification.models import get_online_model
from kolibri.tasks.classification.sklearn_estimator import SklearnEstimator


class StreamedClassifier(SklearnEstimator):
    name = 'streamed_classifier'

    def __init__(self, component_config, classifier='MultinomialNB'):
        super().__init__(component_config=component_config, classifier=classifier)

        self.clf = get_online_model(classifier)

    def partial_fit(self, X, y, classes):
        self.indexer.build_label_dict(y)
        y = self.indexer.numerize_label_sequences(y)

        return self.clf.partial_fit(X, y, classes)



import os
import unittest

from tensorflow.keras.utils import get_file

from kolibri.dnn.embeddings import WordEmbedding
from kolibri.settings import DATA_PATH
from tests.dnn.text.test_embeddings.test_default_embedding import TestDefaultEmbedding

sample_w2v_path = os.path.join(get_file("text",
                                        "https://www.dropbox.com/s/0lducehaevowqpm/test.tar.gz?dl=1",
                                        cache_dir=DATA_PATH,
                                        cache_subdir='test',
                                        untar=True), 'sample_w2v.txt')


class TestWordEmbedding(TestDefaultEmbedding):

    def build_embedding(self):
        embedding = WordEmbedding(sample_w2v_path)
        return embedding


if __name__ == '__main__':
    unittest.main()



import unittest

import tests.dnn.text.test_classification.test_bi_lstm_model as base
from kolibri.dnn.embeddings import WordEmbedding
from kolibri.dnn.tasks.text.classification import BiGRU_Model
from tests.dnn.text.test_embeddings.test_word_embedding import sample_w2v_path


class TestBiGRU_Model(base.TestBiLSTM_Model):
    @classmethod
    def setUpClass(cls):
        cls.EPOCH_COUNT = 2
        cls.TASK_MODEL_CLASS = BiGRU_Model
        cls.w2v_embedding = WordEmbedding(sample_w2v_path)


if __name__ == "__main__":
    unittest.main()



import errno
import io
import json
import os
import pickle

import simplejson


def read_file(filename, encoding="utf-8"):
    """Read text from a file."""
    with io.open(filename, encoding=encoding) as f:
        return f.read()


def read_json_file(filename, encoding="utf-8"):
    """Read json from a file."""
    content = read_file(filename, encoding=encoding)
    try:
        return simplejson.loads(content)
    except ValueError as e:
        raise ValueError("Failed to read json from '{}'. Error: "
                         "{}".format(os.path.abspath(filename), e))


def create_dir(dir_path):
    """Creates a directory and its super paths.

    Succeeds even if the path already exists."""

    try:
        os.makedirs(dir_path)
    except OSError as e:
        # be happy if someone already created the path
        if e.errno != errno.EEXIST:
            raise


def json_to_string(obj, **kwargs):
    indent = kwargs.pop("indent", 2)
    ensure_ascii = kwargs.pop("ensure_ascii", False)
    return json.dumps(obj, indent=indent, ensure_ascii=ensure_ascii, **kwargs)


def write_json_to_file(filename, obj, **kwargs):
    """Write an object as a json string to a file."""

    write_to_file(filename, json_to_string(obj, **kwargs))


def write_to_file(filename, text, encoding='utf-8'):
    """Write a text to a file."""

    with io.open(filename, 'w', encoding=encoding) as f:
        f.write(str(text))


def save_to_disk(path_to_disk, obj, overwrite=False):
    """ Pickle an object to disk """
    dirname = os.path.dirname(path_to_disk)
    if not os.path.exists(dirname):
        raise ValueError("Path " + dirname + " does not exist")

    if not overwrite and os.path.exists(path_to_disk):
        raise ValueError("File " + path_to_disk + "already exists")

    pickle.dump(obj, open(path_to_disk, 'wb'))


def load_from_disk(path_to_disk):
    """ Load a pickle from disk to memory """
    if not os.path.exists(path_to_disk):
        raise ValueError("File " + path_to_disk + " does not exist")

    return pickle.load(open(path_to_disk, 'rb'))



class Token():
    __slots__ = [
        'text',
        'lemma',
        'pos',
        'coarse_pos',
        'data_dic',
        'id',
        'is_stop',
        'is_alpha',
        'isnumeric',
        'islower',
        'istitle',
        'isupper'
    ]



import re
import warnings
from collections import defaultdict

import numpy as np
import scipy.sparse as sp
from sklearn.feature_extraction.text import CountVectorizer, _document_frequency
from sklearn.preprocessing import normalize

FLOAT_DTYPES = (np.float64, np.float32, np.float16)
_white_spaces = re.compile(r"\s\s+")


class StreamedTfidfVectorizer(CountVectorizer):
    """Convert a collection of raw documents to a matrix of TF-IDF features.

    """

    def __init__(self, input='content', encoding='utf-8',
                 decode_error='strict', strip_accents=None, lowercase=True,
                 preindexer=None, tokenizer=None, analyzer='word',
                 stop_words=None, token_pattern=r"(?u)\b\w\w+\b",
                 ngram_range=(1, 1), max_df=1.0, min_df=1,
                 max_features=None, vocabulary=None, binary=False,
                 dtype=np.float64, norm='l2', use_idf=True, smooth_idf=True,
                 sublinear_tf=False):

        super().__init__(
            input=input, encoding=encoding, decode_error=decode_error,
            strip_accents=strip_accents, lowercase=lowercase,
            preindexer=preindexer, tokenizer=tokenizer, analyzer=analyzer,
            stop_words=stop_words, token_pattern=token_pattern,
            ngram_range=ngram_range, max_df=max_df, min_df=min_df,
            max_features=max_features, vocabulary=vocabulary, binary=binary,
            dtype=dtype)

        self.use_idf = use_idf
        self.smooth_idf = smooth_idf
        self.sublinear_tf = sublinear_tf
        self.idf_ = None
        self.norm = norm
        self.vocabulary_ = None
        self._idf_diag = None
        self.n_samples = 0

    # Broadcast the TF-IDF parameters to the underlying transformer instance
    # for easy grid search and repr

    def _check_params(self):
        if self.dtype not in FLOAT_DTYPES:
            warnings.warn("Only {} 'dtype' should be used. {} 'dtype' will "
                          "be converted to np.float64."
                          .format(FLOAT_DTYPES, self.dtype),
                          UserWarning)

    def _fit_tf_idf(self, X):
        if not sp.issparse(X):
            X = sp.csr_matrix(X)
        dtype = X.dtype if X.dtype in FLOAT_DTYPES else np.float64

        if self.use_idf:
            n_samples, n_features = X.shape
            df = _document_frequency(X)
            df = df.astype(dtype)

            # perform idf smoothing if required
            df += int(self.smooth_idf)
            n_samples += int(self.smooth_idf)

            # log+1 instead of log makes sure terms with zero idf don't get
            # suppressed entirely.
            idf = np.log(n_samples / df) + 1
            self._idf_diag = sp.diags(idf, offsets=0,
                                      shape=(n_features, n_features),
                                      format='csr',
                                      dtype=dtype)

    def finalize_partial(self):
        if self.use_idf:
            n_features = len(self._idf_diag)

            self._idf_diag = dict(sorted(self._idf_diag.items()))

            df = np.array(list(self._idf_diag.values()))

            # perform idf smoothing if required
            df += int(self.smooth_idf)
            self.n_samples += int(self.smooth_idf)

            # log+1 instead of log makes sure terms with zero idf don't get
            # suppressed entirely.
            idf = np.log(self.n_samples / df) + 1
            self._idf_diag = sp.diags(idf, offsets=0,
                                      shape=(n_features, n_features),
                                      format='csr')

    def _update_tf_idf(self, raw_documents):
        if not self.vocabulary_:
            self.vocabulary_ = defaultdict()
            self.vocabulary_.default_factory = self.vocabulary_.__len__

        if not self._idf_diag:
            self._idf_diag = {}

        analyze = self.build_analyzer()

        for doc in raw_documents:
            self.n_samples += 1
            features = set(analyze(doc))
            for feature in features:
                try:
                    feature_idx = self.vocabulary_[feature]
                    if feature_idx not in self._idf_diag:
                        self._idf_diag[feature_idx] = 1
                    else:
                        self._idf_diag[feature_idx] += 1
                except KeyError:
                    # Ignore out-of-vocabulary items for fixed_vocab=True
                    continue

    def _transform_tf_idf(self, X):
        if not sp.issparse(X):
            X = sp.csr_matrix(X, dtype=np.float64)

        n_samples, n_features = X.shape

        if self.sublinear_tf:
            np.log(X.data, X.data)
            X.data += 1

        if self.use_idf:
            # idf_ being a property, the automatic attributes detection
            # does not work as usual and we need to specify the attribute
            # name:
            expected_n_features = self._idf_diag.shape[0]
            if n_features != expected_n_features:
                raise ValueError("Input has n_features=%d while the model"
                                 " has been trained with n_features=%d" % (
                                     n_features, expected_n_features))
            # *= doesn't work
            X = X * self._idf_diag

        if self.norm:
            X = normalize(X, norm=self.norm, copy=False)

        return X

    def fit(self, raw_documents, y=None):
        """Learn vocabulary and idf from training set.
        Parameters
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode or file objects.
        y : None
            This parameter is not needed to compute tfidf.
        Returns
        -------
        self : object
            Fitted vectorizer.
        """
        self._check_params()
        X = super().fit_transform(raw_documents)
        self._fit_tf_idf(X)
        return self

    def fit_partial(self, raw_documents, y=None):
        """Learn vocabulary and idf from training set.
        Parameters incrementally
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode or file objects.
        y : None
            This parameter is not needed to compute tfidf.

       Returns
        -------
        self : object
            Fitted vectorizer.
        """

        self._update_tf_idf(raw_documents)
        return self

    def fit_transform(self, raw_documents, y=None):
        """Learn vocabulary and idf, return term-document matrix.
        This is equivalent to fit followed by transform, but more efficiently
        implemented.
        Parameters
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode or file objects.
        y : None
            This parameter is ignored.
        Returns
        -------
        X : sparse matrix, [n_samples, n_features]
            Tf-idf-weighted document-term matrix.
        """
        self._check_params()
        X = super().fit_transform(raw_documents)
        self._fit_tf_idf(X)
        # X is already a transformed view of raw_documents so
        # we set copy to False
        return self._transform_tf_idf(X)

    def transform(self, raw_documents, copy="deprecated"):
        """Transform documents to document-term matrix.
        Uses the vocabulary and document frequencies (df) learned by fit (or
        fit_transform).
        Parameters
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode or file objects.
        copy : bool, default True
            Whether to copy X and operate on the copy or perform in-place
            operations.
            .. deprecated:: 0.22
               The `copy` parameter is unused and was deprecated in version
               0.22 and will be removed in 0.24. This parameter will be
               ignored.
        Returns
        -------
        X : sparse matrix, [n_samples, n_features]
            Tf-idf-weighted document-term matrix.
        """

        # FIXME Remove copy parameter support in 0.24
        if copy != "deprecated":
            msg = ("'copy' param is unused and has been deprecated since "
                   "version 0.22. Backward compatibility for 'copy' will "
                   "be removed in 0.24.")
            warnings.warn(msg, FutureWarning)
        X = super().transform(raw_documents)

        return self._transform_tf_idf(X)

    def _more_tags(self):
        return {'X_types': ['string'], '_skip_test': True}



import json
import logging
import os
import pathlib
from typing import Dict, Any, List, Optional, Union, Tuple

import numpy as np
import tensorflow as tf
from tensorflow import keras

from kolibri.dnn import __version__
from kolibri.dnn.utils import custom_object_scope

L = keras.layers


class BaseAutoencoder(object):
    """Base Sequence Labeling Model"""

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def info(self):
        model_json_str = self.tf_model.to_json()

        return {
            'config': {
                'hyper_parameters': self.hyper_parameters,
            },
            'tf_model': json.loads(model_json_str),
            'class_name': self.__class__.__name__,
            'module': self.__class__.__module__,
            'tf_version': tf.__version__,
            'kolibri.version': __version__
        }

    def __init__(self, hyper_parameters: Optional[Dict[str, Dict[str, Any]]] = None):
        """

        """

        self.hyper_parameters = self.get_default_hyper_parameters()
        self.model_info = {}

        self.autoencoder=None
        if hyper_parameters:
            self.hyper_parameters.update(hyper_parameters)

    def build_model(self, x_train):
        """
        Build model with corpus

        Args:
            x_train: Array of train feature texts (if the model has a single input),
                or tuple of train feature texts array (if the model has multiple inputs)
            y_train: Array of train label texts
            x_validate: Array of validation feature texts (if the model has a single input),
                or tuple of validation feature texts array (if the model has multiple inputs)
            y_validate: Array of validation label texts

        Returns:

        """
        self.input_dim=x_train[0].size


        if self.autoencoder is None:
            self.build_model_arc()
            self.compile_model(**self.hyper_parameters)


    def build_multi_gpu_model(self,
                              gpus: int,
                              x_train: Union[Tuple[List[List[str]], ...], List[List[str]]],
                              y_train: Union[List[List[str]], List[str]],
                              cpu_merge: bool = True,
                              cpu_relocation: bool = False,
                              x_validate: Union[Tuple[List[List[str]], ...], List[List[str]]] = None,
                              y_validate: Union[List[List[str]], List[str]] = None):
        """
        Build multi-GPU model with corpus

        Args:
            gpus: Integer >= 2, number of on GPUs on which to create model replicas.
            cpu_merge: A boolean value to identify whether to force merging model weights
                under the scope of the CPU or not.
            cpu_relocation: A boolean value to identify whether to create the model's weights
                under the scope of the CPU. If the model is not defined under any preceding device
                scope, you can still rescue it by activating this option.
            x_train: Array of train feature texts (if the model has a single input),
                or tuple of train feature texts array (if the model has multiple inputs)
            y_train: Array of train label texts
            x_validate: Array of validation feature texts (if the model has a single input),
                or tuple of validation feature texts array (if the model has multiple inputs)
            y_validate: Array of validation label texts

        Returns:

        """
        if x_validate is not None and not isinstance(x_validate, tuple):
            self.embedding.analyze_corpus(x_train + x_validate, y_train + y_validate)
        else:
            self.embedding.analyze_corpus(x_train, y_train)

        if self.tf_model is None:
            with custom_object_scope():
                self.build_model_arc()
                self.tf_model = tf.keras.utils.multi_gpu_model(self.tf_model,
                                                               gpus,
                                                               cpu_merge=cpu_merge,
                                                               cpu_relocation=cpu_relocation)
                self.compile_model()

    def get_data_generator(self,
                           x_data,
                           y_data,
                           batch_size: int = 64,
                           shuffle: bool = True):
        """
        texts generator for fit_generator

        Args:
            x_data: Array of feature texts (if the model has a single input),
                or tuple of feature texts array (if the model has multiple inputs)
            y_data: Array of label texts
            batch_size: Number of samples per gradient update, default to 64.
            shuffle:

        Returns:
            texts generator
        """
        index_list = np.arange(len(x_data))
        page_count = len(x_data) // batch_size + 1

        while True:
            if shuffle:
                np.random.shuffle(index_list)
            for page in range(page_count):
                start_index = page * batch_size
                end_index = start_index + batch_size
                target_index = index_list[start_index: end_index]

                if len(target_index) == 0:
                    target_index = index_list[0: batch_size]

                yield (x_data, x_data)

    def fit(self,
            x_train: Union[Tuple[List[List[str]], ...], List[List[str]]],
            x_validate: Union[Tuple[List[List[str]], ...], List[List[str]]] = None,
            batch_size: int = 64,
            epochs: int = 5,
            validation_split=0.2,
            callbacks: List[keras.callbacks.Callback] = None,
            fit_kwargs: Dict = None,
            shuffle: bool = True):
        """
        Trains the model for a given number of epochs with fit_generator (iterations on a dataset).

        Args:
            x_train: Array of train feature texts (if the model has a single input),
                or tuple of train feature texts array (if the model has multiple inputs)
            y_train: Array of train label texts
            x_validate: Array of validation feature texts (if the model has a single input),
                or tuple of validation feature texts array (if the model has multiple inputs)
            y_validate: Array of validation label texts
            batch_size: Number of samples per gradient update, default to 64.
            epochs: Integer. Number of epochs to train the model. default 5.
            callbacks:
            fit_kwargs: fit_kwargs: additional arguments passed to ``fit_generator()`` function from
                ``tensorflow.keras.Model``
                - https://www.tensorflow.org/api_docs/python/tf/keras/models/Model#fit_generator
            shuffle:

        Returns:

        """
        self.build_model(x_train)

        with custom_object_scope():

            return self.autoencoder.fit(x_train, x_train, epochs=epochs,
                                        batch_size=batch_size,
                                        shuffle=shuffle, validation_split=validation_split
                                        )

    def compile_model(self, **kwargs):
        """Configures the model for training.

        Using ``compile()`` function of ``tf.keras.Model`` -
        https://www.tensorflow.org/api_docs/python/tf/keras/models/Model#compile

        Args:
            **kwargs: arguments passed to ``compile()`` function of ``tf.keras.Model``

        Defaults:
            - loss: ``categorical_crossentropy``
            - optimizer: ``adam``
            - metrics: ``['accuracy']``
        """
        configs={}
        configs['loss']=kwargs.get('loss', 'mean_squared_error')
        configs['optimizer']=kwargs.get('optimizer', 'adam')
#        configs['metrics']=kwargs.get('metrics', ['loss'])


        self.autoencoder.compile(**configs)

    def encode(self,
                x_data,
                batch_size=32,
                debug_info=False,
                predict_kwargs: Dict = None):
        """
        Generates output predictions for the encoder.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with custom_object_scope():
#            tensor = self.process_x_dataset(x_data)
            pred = self.encoder.predict(x_data, batch_size=batch_size, **predict_kwargs)



            if debug_info:
                print('input: {}'.format(x_data))
                print('output: {}'.format(pred))

        return pred

    def decode(self,
               x_data,
               batch_size=32,
               debug_info=False,
               predict_kwargs: Dict = None):
        """
        Generates output predictions for the encoder.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with custom_object_scope():
            tensor = self.process_x_dataset(x_data)
            pred = self.decoder.predict(tensor, batch_size=batch_size, **predict_kwargs)

            if debug_info:
                print('input: {}'.format(tensor))
                print('output: {}'.format(pred))

        return pred

    def predict(self,
               x_data,
               batch_size=32,
               debug_info=False,
               predict_kwargs: Dict = None):
        """
        Generates output predictions for the encoder.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with custom_object_scope():

            pred = self.autoencoder.predict(x_data, batch_size=batch_size, **predict_kwargs)

            if debug_info:
                print('input: {}'.format(x_data))
                print('output: {}'.format(pred))

        return pred

    def evaluate(self,
                 x_data,
                 batch_size=None,
                 digits=4,
                 output_dict=False,
                 debug_info=False) -> Optional[Tuple[float, float, Dict]]:
        x_pred = self.autoencoder.predict(x_data, batch_size=batch_size)

        loss=keras.losses.mean_squared_error(x_data, x_pred)
        print(tf.math.reduce_mean(
            loss, axis=None, keepdims=False, name=None
        ))



    def summary(self):
        self.autoencoder.summary()

    def build_model_arc(self):
        raise NotImplementedError

    def save(self, model_path: str):
        """
        Save model
        Args:
            model_path:

        Returns:

        """
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)

        with open(os.path.join(model_path, 'model_info.json'), 'w') as f:
            f.write(json.dumps(self.info(), indent=2, ensure_ascii=True))
            f.close()

        self.tf_model.save_weights(os.path.join(model_path, 'model_weights.h5'))
        logging.info('model saved to {}'.format(os.path.abspath(model_path)))

    def save_info(self, model_path: str):
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)

        with open(os.path.join(model_path, 'model_info.json'), 'w') as f:
            f.write(json.dumps(self.info(), indent=2, ensure_ascii=True))
            f.close()




# from tensorflow.keras.callbacks import EarlyStopping, TensorBoard
#
#
# class BaseAutoencoder(object):
#
#     def __index__(self, log_dir='/tmp/autoencoder', stop_early=True):
#         self.autoencoder = None
#         self.encoder = None
#         self.decoder = None
#         self.callbacks = []
#         if log_dir:
#             self.callbacks.append(TensorBoard(log_dir=log_dir))
#         if stop_early:
#             self.callbacks.append(EarlyStopping(monitor='val_loss', patience=2, verbose=1, mode='auto'))
#
#     def fit(self, x_train, x_test=None, epochs=5, batch_size=32):
#
#         test_data = None
#         if x_test:
#             test_data = (x_test, x_test)
#
#         self.autoencoder.fit(x_train, x_train,
#                              nb_epoch=epochs,
#                              batch_size=batch_size,
#                              shuffle=True,
#                              validation_data=test_data,
#                              callbacks=self.callbacks)
#
#     def encode(self, x):
#         return self.encoder.predict(x)
#
#     def decode(self, x):
#         return self.decoder.predict(x)
#
#     def summary(self):
#         self.autoencoder.summary()



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.layers import TimeDistributed
from tensorflow.keras.models import Model
from tensorflow.keras.regularizers import l2

from kolibri.dnn.layers import L
from kolibri.dnn.layers import Melspectrogram, Normalization2D
from kolibri.dnn.tasks.audio.base_model import BaseAudioClassificationModel


class Conv1D_Model(BaseAudioClassificationModel):

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'delta_time': 1.0,
            'layer_spectrogram': {
                'n_dft': 512,
                'n_hop': 160,
                'padding': 'same',
                'sr': 16000,
                'n_mels': 128,
                'fmin': 0.0,
                'fmax': 8000,
                'power_melgram': 2.0,
                'return_decibel_melgram': True,
                'trainable_fb': False,
                'trainable_kernel': False,
                'name': 'melbands'
            },
            'layer_dense': {
                'activation': 'softmax',
                'name': 'softmax'
            },
            'max_pool_2D_layer': {
                'pool_size': (2, 2)
            },
            'layer_time_distributed': {
                'filters': [8, 16, 32, 64, 128],
                'kernel_size': 4,
                'activation': 'relu',
                'padding': 'same'
            },
            'dense_layer': {

                'activity_regularizer': l2(0.001),
                'units': 64,
                'activation': 'relu',
            },
            'activation_layer': {
                'activation': 'softmax'
            },
            'multipooling': False,
            'dropout': 0.2
        }

    def build_model_arc(self):

        output_dim = self.label_indexer.vocab_size
        config = self.hyper_parameters

        filters = config['layer_time_distributed']['filters']

        i = L.Input(shape=(1, int(config['layer_spectrogram']['sr']) * int(config['delta_time'])), name='input')
        layer_mel_spectrogram = Melspectrogram(**config['layer_spectrogram'])(i)
        norm = Normalization2D(str_axis='batch', name='batch_norm')(layer_mel_spectrogram)
        permute = L.Permute((2, 1, 3), name='permute')(norm)

        # build model structure in sequent way
        filter_id = 0
        for filter_size in filters:
            if filter_id == 0:
                x = L.TimeDistributed(L.Conv1D(
                    filter_size,
                    kernel_size=config['layer_time_distributed']['kernel_size'],
                    padding=config['layer_time_distributed']['padding'],
                    activation=config['layer_time_distributed']['activation']))(permute)
                filter_id += 1
            else:
                x = L.TimeDistributed(L.Conv1D(
                    filter_size,
                    kernel_size=config['layer_time_distributed']['kernel_size'],
                    padding=config['layer_time_distributed']['padding'],
                    activation=config['layer_time_distributed']['activation']))(x)
                filter_id += 1
            if config['multipooling'] and filter_id < len(filters):
                x = L.MaxPooling2D(**config['max_pool_2D_layer'])(x)

        x = L.GlobalMaxPooling2D()(x)
        if config['dropout'] > 0:
            x = L.Dropout(config['dropout'])(x)
        #        x = L.Dense(64, activation='relu', activity_regularizer=l2(0.001), name='dense')(x)
        x = L.Dense(**config['dense_layer'])(x)
        tensor = L.Dense(output_dim, **config['activation_layer'])(x)

        self.tf_model = tf.keras.Model(i, tensor)


def Conv1D(N_CLASSES=10, SR=16000, DT=1.0):
    i = layers.Input(shape=(1, int(SR * DT)), name='input')
    x = Melspectrogram(n_dft=512, n_hop=160,
                       padding='same', sr=SR, n_mels=128,
                       fmin=0.0, fmax=SR / 2, power_melgram=2.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False,
                       name='melbands')(i)
    x = Normalization2D(str_axis='batch', name='batch_norm')(x)
    x = layers.Permute((2, 1, 3), name='permute')(x)
    x = TimeDistributed(layers.Conv1D(8, kernel_size=(4), activation='tanh'), name='td_conv_1d_tanh')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), name='max_pool_2d_1')(x)
    x = TimeDistributed(layers.Conv1D(16, kernel_size=(4), activation='relu'), name='td_conv_1d_relu_1')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), name='max_pool_2d_2')(x)
    x = TimeDistributed(layers.Conv1D(32, kernel_size=(4), activation='relu'), name='td_conv_1d_relu_2')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), name='max_pool_2d_3')(x)
    x = TimeDistributed(layers.Conv1D(64, kernel_size=(4), activation='relu'), name='td_conv_1d_relu_3')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), name='max_pool_2d_4')(x)
    x = TimeDistributed(layers.Conv1D(128, kernel_size=(4), activation='relu'), name='td_conv_1d_relu_4')(x)
    x = layers.GlobalMaxPooling2D(name='global_max_pooling_2d')(x)
    x = layers.Dropout(rate=0.1, name='dropout')(x)
    x = layers.Dense(64, activation='relu', activity_regularizer=l2(0.001), name='dense')(x)
    o = layers.Dense(N_CLASSES, activation='softmax', name='softmax')(x)

    model = Model(inputs=i, outputs=o, name='1d_convolution')
    model.compile(optimizer='adam',
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])

    return model


def Conv2D(N_CLASSES=10, SR=16000, DT=1.0):
    i = layers.Input(shape=(1, int(SR * DT)), name='input')
    x = Melspectrogram(n_dft=512, n_hop=160,
                       padding='same', sr=SR, n_mels=128,
                       fmin=0.0, fmax=SR / 2, power_melgram=2.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False,
                       name='melbands')(i)
    x = Normalization2D(str_axis='batch', name='batch_norm')(x)
    x = layers.Conv2D(8, kernel_size=(7, 7), activation='tanh', padding='same', name='conv2d_tanh')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), padding='same', name='max_pool_2d_1')(x)
    x = layers.Conv2D(16, kernel_size=(5, 5), activation='relu', padding='same', name='conv2d_relu_1')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), padding='same', name='max_pool_2d_2')(x)
    x = layers.Conv2D(16, kernel_size=(3, 3), activation='relu', padding='same', name='conv2d_relu_2')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), padding='same', name='max_pool_2d_3')(x)
    x = layers.Conv2D(32, kernel_size=(3, 3), activation='relu', padding='same', name='conv2d_relu_3')(x)
    x = layers.MaxPooling2D(pool_size=(2, 2), padding='same', name='max_pool_2d_4')(x)
    x = layers.Conv2D(32, kernel_size=(3, 3), activation='relu', padding='same', name='conv2d_relu_4')(x)
    x = layers.Flatten(name='flatten')(x)
    x = layers.Dropout(rate=0.2, name='dropout')(x)
    x = layers.Dense(64, activation='relu', activity_regularizer=l2(0.001), name='dense')(x)
    o = layers.Dense(N_CLASSES, activation='softmax', name='softmax')(x)

    model = Model(inputs=i, outputs=o, name='2d_convolution')
    model.compile(optimizer='adam',
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])

    return model


def LSTM(N_CLASSES=10, SR=16000, DT=1.0):
    i = layers.Input(shape=(1, int(SR * DT)), name='input')
    x = Melspectrogram(n_dft=512, n_hop=160,
                       padding='same', sr=SR, n_mels=128,
                       fmin=0.0, fmax=SR / 2, power_melgram=2.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False,
                       name='melbands')(i)
    x = Normalization2D(str_axis='batch', name='batch_norm')(x)
    x = layers.Permute((2, 1, 3), name='permute')(x)
    x = TimeDistributed(layers.Reshape((-1,)), name='reshape')(x)
    s = TimeDistributed(layers.Dense(64, activation='tanh'),
                        name='td_dense_tanh')(x)
    x = layers.Bidirectional(layers.LSTM(32, return_sequences=True),
                             name='bidirectional_lstm')(s)
    x = layers.concatenate([s, x], axis=2, name='skip_connection')
    x = layers.Dense(64, activation='relu', name='dense_1_relu')(x)
    x = layers.MaxPooling1D(name='max_pool_1d')(x)
    x = layers.Dense(32, activation='relu', name='dense_2_relu')(x)
    x = layers.Flatten(name='flatten')(x)
    x = layers.Dropout(rate=0.2, name='dropout')(x)
    x = layers.Dense(32, activation='relu',
                     activity_regularizer=l2(0.001),
                     name='dense_3_relu')(x)
    o = layers.Dense(N_CLASSES, activation='softmax', name='softmax')(x)

    model = Model(inputs=i, outputs=o, name='long_short_term_memory')
    model.compile(optimizer='adam',
                  loss='categorical_crossentropy',
                  metrics=['accuracy'])

    return model


def get_model(model_type, hyper_parameters=None):
    if model_type == 'conv_1d':
        return Conv1D_Model(hyper_parameters=hyper_parameters)
    else:
        raise Exception('Model does not exist in the library of audio classification models')



import os
from glob import glob
from typing import Any
from typing import Iterable, Iterator

import numpy as np
import tensorflow as tf
from tensorflow import keras


class DataGenerator(keras.utils.Sequence):

    def __init__(self, x_values, y_values, content_indexer=None, label_indexer=None, x_augmentor=None, batch_size=1,
                 shuffle=False):
        self.y_values = y_values  # array of y_values
        self.x_values = x_values  # array of image paths
        self.batch_size = batch_size  # batch size
        self.shuffle = shuffle  # shuffle bool
        self.on_epoch_end()
        self.content_indexer = content_indexer
        self.label_indexer = label_indexer
        self.augmentor = x_augmentor

    def __len__(self):
        'Denotes the number of batches per epoch'
        if self.x_values is not None:
            return int(np.floor(len(self.x_values) / self.batch_size))
        else:
            return int(np.floor(len(self.y_values) / self.batch_size))

    def set_batch_size(self, new_batch_size):
        self.batch_size = new_batch_size

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        if self.x_values is not None:
            self.indexes = np.arange(len(self.x_values))
        else:
            self.indexes = np.arange(len(self.y_values))

        if self.shuffle:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        'Generate one batch of texts'
        # selects indices of texts for next batch
        indexes = self.indexes[index * self.batch_size: (index + 1) * self.batch_size]

        # select texts and load images
        labels = None
        content = None

        if self.y_values is not None:
            labels = np.array([self.y_values[k] for k in indexes])
            if self.label_indexer:
                labels = self.label_indexer.transform(labels)
            if self.batch_size == 1:
                labels = labels[0]
        if self.x_values is not None:
            # preprocess and augment texts

            content = [self.x_values[k] for k in indexes]
            if self.augmentor:
                content = self.augmentor(content)

            if self.content_indexer:
                content = self.content_indexer.transform(content)

            if self.batch_size == 1:
                content = content[0]
            content = np.array(content)
        return content, labels

    def get_data(self):
        return [x for x, y in self], [y for x, y in self]


class FolderDataGenerator(DataGenerator):

    def __init__(self, folder_path, content_indexer=None, label_indexer=None, x_augmentor=None,
                 batch_size=64, shuffle=False):

        file_paths = glob('{}/**'.format(folder_path), recursive=True)
        file_paths = [x.replace(os.sep, '/') for x in file_paths if os.path.isfile(x)]
        labels = [os.path.basename(os.path.dirname(x)) for x in file_paths]


        super().__init__(x_values=file_paths, y_values=labels, content_indexer=content_indexer,
                         label_indexer=label_indexer, x_augmentor=x_augmentor, batch_size=batch_size, shuffle=shuffle)



class Seq2SeqDataSet(Iterable):
    def __init__(self,
                 corpus,
                 *,
                 batch_size: int = 64,
                 encoder_processor,
                 decoder_processor,
                 encoder_seq_length: int = None,
                 decoder_seq_length: int = None):
        self.corpus = corpus

        self.encoder_processor = encoder_processor
        self.decoder_processor = decoder_processor

        self.encoder_seq_length = encoder_seq_length
        self.decoder_seq_length = decoder_seq_length

        self.batch_size = batch_size

    def __len__(self) -> int:
        return max(len(self.corpus) // self.batch_size, 1)

    def __iter__(self) -> Iterator:
        batch_x, batch_y = [], []
        for x, y in self.corpus:
            batch_x.append(x)
            batch_y.append(y)
            if len(batch_x) == self.batch_size:
                x_tensor = self.encoder_processor.transform(batch_x, )
                y_tensor = self.decoder_processor.transform(batch_y, )
                yield x_tensor, y_tensor
                batch_x, batch_y = [], []

    def items(self, batch_count: int = None) -> Any:
        x_shape = [self.batch_size, self.encoder_seq_length]
        y_shape = [self.batch_size, self.decoder_seq_length]
        dataset = tf.data.Dataset.from_generator(self.__iter__,
                                                 output_types=(tf.int64, tf.int64),
                                                 output_shapes=(x_shape, y_shape))
        dataset = dataset.repeat()
        dataset = dataset.prefetch(50)
        if batch_count is None:
            batch_count = len(self)
        return dataset.take(batch_count)



import unittest

from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus
from kolibri.dnn.utils import load_data_object
from kolibri.indexers import LabelIndexer
from kolibri.tokenizer import WordTokenizer

tokenizer = WordTokenizer()


class TestClassificationProcessor(unittest.TestCase):
    def test_indexer(self):
        corpus = ConsumerComplaintsCorpus()
        x_set, y_set = corpus.get_data(nb_samples=500)
        indexer = LabelIndexer()
        indexer.build_vocab(x_set, y_set)
        transformed_idx = indexer.transform(y_set[20:40], )

        info_dict = indexer.to_dict()

        p2: LabelIndexer = load_data_object(info_dict)
        assert (transformed_idx == p2.transform(y_set[20:40], )).all()
        assert y_set[20:40] == p2.inverse_transform(transformed_idx)

    # def test_multi_label_processor(self):
    #     from kolibri.texts.corpus import ConsumerComplaintsCorpus
    #     corpus=ConsumerComplaintsCorpus()
    #     x_set, y_set = corpus.get_data(nb_samples=500)
    #     x_set=tokenizer.tokenizer(x_set)
    #     corpus_gen = CorpusGenerator(x_set, y_set)
    #
    #     indexer = LabelIndexer(multi_label=True)
    #     indexer.build_vocab_generator([corpus_gen])
    #     transformed_idx = indexer.transform(y_set[20:40])
    #
    #     info_dict = indexer.to_dict()
    #
    #     p2: LabelIndexer = load_data_object(info_dict)
    #     assert (transformed_idx == p2.transform(y_set[20:40])).all()
    #
    #     x1s = y_set[20:40]
    #     x2s = p2.inverse_transform(transformed_idx)
    #     for sample_x1, sample_x2 in zip(x1s, x2s):
    #         assert sorted(sample_x1) == sorted(sample_x2)


if __name__ == "__main__":
    pass



# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: stacked_embedding.py
# time: 2019-05-23 09:18

import json
import pydoc
from typing import Union, Optional, Tuple, List, Dict

import numpy as np
import tensorflow as tf
from tensorflow.python import keras

import kolibri.dnn
from kolibri.dnn.embeddings.base_embedding import Embedding
from kolibri.dnn.layers import L


# from kolibri.indexers import BaseIndexer


class StackedEmbedding(Embedding):
    """Embedding layer without pre-training, train embedding layer while training model"""

    @classmethod
    def _load_saved_instance(cls,
                             config_dict: Dict,
                             model_path: str,
                             tf_model: keras.Model):
        embeddings = []
        for embed_info in config_dict['embeddings']:
            embed_class = pydoc.locate(f"{embed_info['module']}.{embed_info['class_name']}")
            embedding: Embedding = embed_class._load_saved_instance(embed_info,
                                                                    model_path,
                                                                    tf_model)
            embeddings.append(embedding)
        instance = cls(embeddings=embeddings,
                       from_saved_model=True)
        print('----')
        print(instance.embeddings)

        embed_model_json_str = json.dumps(config_dict['embed_model'])
        instance.embed_model = keras.models.model_from_json(embed_model_json_str,
                                                            custom_objects=kolibri.dnn.custom_objects)
        # Load Weights from model
        for layer in instance.embed_model.layers:
            layer.set_weights(tf_model.get_layer(layer.my_name).get_weights())
        return instance

    def info(self):
        info = super(StackedEmbedding, self).info()
        info['embeddings'] = [embed.info() for embed in self.embeddings]
        info['config'] = {}
        return info

    def __init__(self,
                 embeddings: List[Embedding],
                 indexer=None,
                 from_saved_model: bool = False):
        """

        Args:
            embeddings:
            indexer:
        """
        task = kolibri.dnn.CLASSIFICATION
        if all(isinstance(embed.sequence_length, int) for embed in embeddings):
            sequence_length = [embed.sequence_length for embed in embeddings]
        else:
            raise ValueError('Need to set sequence length_train for all embeddings while using stacked embedding')

        super(StackedEmbedding, self).__init__(task=task,
                                               sequence_length=sequence_length[0],
                                               embedding_size=100,
                                               indexer=indexer,
                                               from_saved_model=from_saved_model)

        self.embeddings = embeddings
        self.indexer = embeddings[0].indexer

        if not from_saved_model:
            self._build_model()

    def _build_model(self, **kwargs):
        if self.embed_model is None and all(embed.embed_model is not None for embed in self.embeddings):
            layer_concatenate = L.Concatenate(name='layer_concatenate')

            inputs = []

            for embed in self.embeddings:
                inputs += embed.embed_model.inputs

            # inputs = [embed.embed_model.inputs for embed in self.embeddings]
            outputs = layer_concatenate([embed.embed_model.output for embed in self.embeddings])

            self.embed_model = tf.keras.Model(inputs, outputs)

    def analyze_corpus(self,
                       x: Union[Tuple[List[List[str]], ...], List[List[str]]],
                       y: Union[List[List[str]], List[str]]):
        for index in range(len(x)):
            self.embeddings[index].analyze_corpus(x[index], y)
        self._build_model()

    def process_x_dataset(self,
                          data: Tuple[List[List[str]], ...],
                          subset: Optional[List[int]] = None) -> Tuple[np.ndarray, ...]:
        """
        batch process feature texts while training

        Args:
            data: label dataset
            subset: subset index list

        Returns:
            vectorized feature tensor
        """
        result = []
        for index, dataset in enumerate(data):
            x = self.embeddings[index].process_x_dataset(dataset, subset)
            if isinstance(x, tuple):
                result += list(x)
            else:
                result.append(x)
        return tuple(result)

    def process_y_dataset(self,
                          data: List[List[str]],
                          subset: Optional[List[int]] = None) -> np.ndarray:
        return self.embeddings[0].process_y_dataset(data, subset)


if __name__ == "__main__":
    pass



import os

from kolibri import Component
from kolibri.document import Document
from kolibri.settings import EMBEDDING_SIZE
from kolibri.tools.word2vec import train_word2vec, fit_scaler
from kolibri.utils.file import save_to_disk, load_from_disk

W2V_FILENAME = "word2vec.model"
SCALER_FILENAME = "scaler.model"


class EmbeddingTrainer(Component):
    name = "embedding_trainer"

    provides = ["embedding"]
    defaults = {
        "embedding_path": None,
        "embeddings_dim": EMBEDDING_SIZE
    }

    def __init__(self, config={}):
        super().__init__(config)
        self.embedding_size = self.component_config["embeddings_dim"]

    def train(self, training_data, **kwargs):

        self.train_word2vec(training_data, self.embedding_size)
        self.fit_scaler(training_data)
        self.save_scaler()
        self.save_word2vec_model()

    def transform(self, data, **kwargs):

        return

    def process(self, document: Document, **kwargs):
        return

    def train_word2vec(self, trainingdata, vec_dim=EMBEDDING_SIZE):
        """
        Train the word2vec model on a directory with text files.
        :param train_dir: directory with '.txt' files
        :param vec_dim: dimensionality of the word vectors

        :return: trained gensim model
        """

        self.word2vec_model = train_word2vec(trainingdata, vec_dim=vec_dim)

        return self.word2vec_model

    def fit_scaler(self, train_data):
        """
        Fit a scaler on given texts. Word vectors must be trained already.
        :param train_dir: directory with '.txt' files

        :return: fitted scaler object
        """
        if not self.word2vec_model:
            raise ValueError('word2vec model is not trained. ' + \
                             'Run train_word2vec() first.')

        self.scaler = fit_scaler(train_data, word2vec_model=self.word2vec_model)

        return self.scaler

    def save_scaler(self, overwrite=True):
        """ Save the scaler object to a file """
        if not self.scaler:
            raise ValueError("Can't save the scaler, " + \
                             "it has not been trained yet")
        save_to_disk(os.path.join(self.component_config['output-folder'], SCALER_FILENAME), self.scaler,
                     overwrite=overwrite)

    def load_scaler(self):
        """ Load the scaler object from a file """
        self.scaler = load_from_disk(os.path.join(self.component_config['output-folder'], SCALER_FILENAME))

    def save_word2vec_model(self, overwrite=True):
        """ Save the word2vec model to a file """
        if not self.word2vec_model:
            raise ValueError("Can't save the word2vec model, " + \
                             "it has not been trained yet")
        save_to_disk(os.path.join(self.component_config["output-folder"], W2V_FILENAME), self.word2vec_model,
                     overwrite=overwrite)

    def load_word2vec_model(self):
        """ Load the word2vec model from a file """
        self.word2vec_model = load_from_disk(os.path.join(self.component_config["embedding_path"], W2V_FILENAME))
        return self.word2vec_model

    def get_info(self):
        return "word_tokenizer"



"""
Tokenizer Interface
"""

from kolibri.kolibri_component import Component


class Tokenizer(Component):
    def __init__(self, config={}):
        super().__init__(config)
        self.tokenizer = None

    def tokenize(self, text):
        raise NotImplementedError



import logging
import os
import pathlib
import shutil
import tempfile
from unittest import TestCase

TEST_DIR = tempfile.mkdtemp(prefix="kolibri_tests")


class KolibriTestCase(TestCase):
    """
    A custom subclass of :class:`~unittest.TestCase` that disables some of the
    more verbose AllenNLP logging and that creates and destroys a temp directory
    as a test fixture.
    """

    PROJECT_ROOT = (pathlib.Path(__file__).parent / "..").resolve()
    MODULE_ROOT = PROJECT_ROOT / "kolibri"
    TESTS_ROOT = MODULE_ROOT / "tests"

    def setUp(self):
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s", level=logging.DEBUG
        )

        self.TEST_DIR = pathlib.Path(TEST_DIR)

        os.makedirs(self.TEST_DIR, exist_ok=True)

    def tearDown(self):
        shutil.rmtree(self.TEST_DIR)



import os
import re
from pathlib import Path

from kolibri.utils.dict import Dict

DATA_PATH = os.path.join(str(Path.home()), '.kolibri')

Path(DATA_PATH).mkdir(exist_ok=True, parents=True)

data = Dict({
    'MIN_EXAMPLES_PER_CLASS': 15,
    'MIN_EXAMPLES_PER_ENTITY': 2
})

# [entity_text](entity_type(:entity_synonym)?)
ent_regex = re.compile(r'\[(?P<entity_text>[^\]]+)'
                       r'\]\((?P<entity>[^:)]*?)'
                       r'(?:\:(?P<text>[^)]+))?\)')

DEFAULT_SERVER_PORT = 5005

EMBEDDING_SIZE = 100
WORD2VEC_WORKERS = 4
MIN_WORD_COUNT = 5
WORD2VEC_CONTEXT = 5

MINIMUM_COMPATIBLE_VERSION = "0.0.1"

DEFAULT_NLU_FALLBACK_THRESHOLD = 0.0

DEFAULT_CORE_FALLBACK_THRESHOLD = 0.0

DEFAULT_REQUEST_TIMEOUT = 60 * 5  # 5 minutes

DEFAULT_SERVER_FORMAT = "http://localhost:{}"

DEFAULT_SERVER_URL = DEFAULT_SERVER_FORMAT.format(DEFAULT_SERVER_PORT)

corenlp = Dict({
    'LIBRARY_PATH': '/Users/mbenhaddou/Stanford/stanford-corenlp-full-2018-02-27/',
    'PORT': 9000,
    'D2V_MODEL_PATH': "/Users/mbenhaddou/Documents/Mentis/Developement/Python/Kolibri/kolibri/features/doc2vec/model"
})

modeling = Dict({
    # How many classes are at max put into the output class ranking, everything else will be cut off
    'TARGET_RANKING_LENGTH': 10
})

dialog = Dict({
    'INTENT_MESSAGE_PREFIX': "/",
    'DEFAULT_FALLBACK_ACTION': "action_default_fallback",
    'USER_INTENT_RESTART': "/restart",  # dINTENT_MESSAGE_PREFIX +
    'DOCS_BASE_URL': '.',
    'FALLBACK_SCORE': 1.1
})

DOCUMENT_TEXT_MAX_LENGTH = 255

DOCUMENT_LABEL_MAX_LENGTH = 32

# The maximum length_train of characters that the name of a tag can contain
TAG_NAME_MAX_LENGTH = 50



# encoding: utf-8

# author: Alex
# contact: ialexwwang@gmail.com
# version: 0.1
# license: Apache Licence
# file: dpcnn_model.py
# time: 2019-07-02 19:15
# Reference:
# https://ai.tencent.com/ailab/media/publications/ACL3-Brady.pdf
# https://github.com/Cheneng/DPCNN
# https://github.com/miracleyoo/DPCNN-TextCNN-Pytorch-Inception
# https://www.kaggle.com/michaelsnell/conv1d-dpcnn-in-keras

from math import log2, floor
from typing import Dict, Any

import tensorflow as tf

from kolibri.dnn.layers import L, KMaxPoolingLayer
from kolibri.dnn.tasks.text.classification.base_model import BaseTextClassificationModel


class DPCNN_Model(BaseTextClassificationModel):
    '''
    This implementation of DPCNN requires a clear declared sequence length_train.
    So sequences input in should be padded or cut to a given length_train in advance.
    '''

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        pool_type = 'max'
        filters = 250
        activation = 'linear'
        return {
            'region_embedding': {
                'filters': filters,
                'kernel_size': 3,
                'strides': 1,
                'padding': 'same',
                'activation': activation,
                'name': 'region_embedding',
            },
            'region_dropout': {
                'rate': 0.2,
            },
            'conv_block': {
                'filters': filters,
                'kernel_size': 3,
                'activation': activation,
                'shortcut': True,
            },
            'resnet_block': {
                'filters': filters,
                'kernel_size': 3,
                'activation': activation,
                'shortcut': True,
                'pool_type': pool_type,
                'sorted': True,
            },
            'dense': {
                'units': 256,
                'activation': activation,
            },
            'dropout': 0.5,
            'activation': {
                'activation': 'softmax',
            }
        }

    def downsample(self, inputs, pool_type: str = 'max',
                   sorted: bool = True, stage: int = 1):  # noqa: A002
        layers_pool = []
        if pool_type == 'max':
            layers_pool.append(
                L.MaxPooling1D(pool_size=3,
                               strides=2,
                               padding='same',
                               name=f'pool_{stage}'))
        elif pool_type == 'k_max':
            k = int(inputs.shape[1].value / 2)
            layers_pool.append(
                KMaxPoolingLayer(k=k,
                                 sorted=sorted,
                                 name=f'pool_{stage}'))
        elif pool_type == 'conv':
            layers_pool.append(
                L.Conv1D(filters=inputs.shape[-1].value,
                         kernel_size=3,
                         strides=2,
                         padding='same',
                         name=f'pool_{stage}'))
            layers_pool.append(
                L.BatchNormalization())
        elif pool_type is None:
            layers_pool = []
        else:
            raise ValueError(f'unsupported pooling type `{pool_type}`!')

        tensor_out = inputs
        for layer in layers_pool:
            tensor_out = layer(tensor_out)
        return tensor_out

    def conv_block(self, inputs, filters: int, kernel_size: int = 3,
                   activation: str = 'linear', shortcut: bool = True):
        layers_conv_unit = []
        layers_conv_unit.append(
            L.BatchNormalization())
        layers_conv_unit.append(
            L.PReLU())
        layers_conv_unit.append(
            L.Conv1D(filters=filters,
                     kernel_size=kernel_size,
                     strides=1,
                     padding='same',
                     activation=activation))
        layers_conv_block = layers_conv_unit * 2

        tensor_out = inputs
        for layer in layers_conv_block:
            tensor_out = layer(tensor_out)

        if shortcut:
            tensor_out = L.Add()([inputs, tensor_out])

        return tensor_out

    def resnet_block(self, inputs, filters: int, kernel_size: int = 3,
                     activation: str = 'linear', shortcut: bool = True,
                     pool_type: str = 'max', sorted: bool = True, stage: int = 1):  # noqa: A002
        tensor_pool = self.downsample(inputs, pool_type=pool_type, sorted=sorted, stage=stage)
        tensor_out = self.conv_block(tensor_pool, filters=filters, kernel_size=kernel_size,
                                     activation=activation, shortcut=shortcut)
        return tensor_out

    def build_model_arc(self):
        output_dim = len(self.indexer.label2idx)
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_region = [
            L.Conv1D(**config['region_embedding']),
            L.BatchNormalization(),
            L.PReLU(),
            L.Dropout(**config['region_dropout'])
        ]

        layers_main = [
            L.GlobalMaxPooling1D(),
            L.Dense(**config['dense']),
            L.BatchNormalization(),
            L.PReLU(),
            L.Dropout(config['dropout']),
            L.Dense(output_dim, **config['activation'])
        ]

        tensor_out = embed_model.output

        # build region tensors
        for layer in layers_region:
            tensor_out = layer(tensor_out)

        # build the base pyramid layer
        tensor_out = self.conv_block(tensor_out, **config['conv_block'])
        # build the above pyramid layers while `steps > 2`
        seq_len = tensor_out.shape[1]
        if seq_len is None:
            raise ValueError('`sequence_length` should be explicitly assigned, but it is `None`.')
        for i in range(floor(log2(seq_len)) - 2):
            tensor_out = self.resnet_block(tensor_out, stage=i + 1,
                                           **config['resnet_block'])
        for layer in layers_main:
            tensor_out = layer(tensor_out)

        self.tf_model = tf.keras.Model(embed_model.inputs, tensor_out)



import os
import tempfile
import time
import unittest
from typing import Type

from kolibri.dnn.embeddings import WordEmbedding
from kolibri.dnn.tasks.text.labeling import BiLSTM_Model, BaseLabelingModel
from tests.dnn.text.test_embeddings.test_word_embedding import sample_w2v_path


class TestBiLSTM_Model(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.EPOCH_COUNT = 1
        cls.TASK_MODEL_CLASS: Type[BaseLabelingModel] = BiLSTM_Model

    def test_basic_use(self):
        model = self.TASK_MODEL_CLASS()
        from kolibri.data.corpus.corpus import CONLL2003ENCorpus
        corpus = CONLL2003ENCorpus()

        train_x, train_y = corpus.get_data(nb_samples=500)

        model.fit(train_x,
                  train_y,
                  epochs=self.EPOCH_COUNT)

        model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
        original_y = model.predict(train_x[:20])
        model.save(model_path)
        del model

        new_model = self.TASK_MODEL_CLASS.load_model(model_path)
        new_model.tf_model.summary()

        new_y = new_model.predict(train_x[:20])
        assert new_y == original_y

        report = new_model.evaluate(train_x, train_y)
        print(report)

    def test_with_word_embedding(self):
        w2v_embedding = WordEmbedding(sample_w2v_path)
        model = self.TASK_MODEL_CLASS(embedding=w2v_embedding, sequence_length=120)
        from kolibri.data.corpus.corpus import CONLL2003ENCorpus
        corpus = CONLL2003ENCorpus()

        train_x, train_y = corpus.get_data(nb_samples=500)
        valid_x, valid_y = train_x, train_y

        model.fit(train_x,
                  train_y,
                  x_validate=valid_x,
                  y_validate=valid_y,
                  epochs=self.EPOCH_COUNT)

    # def test_with_bert(self):
    #     bert_path = get_file('bert_sample_model',
    #                          "http://dropbox/bert_sample_model.tar.bz2",
    #                          cache_dir=DATA_PATH,
    #                          untar=True)
    #     embedding = BertEmbedding(model_folder=bert_path)
    #     model = self.TASK_MODEL_CLASS(embedding=embedding)
    #     train_x, train_y = TestMacros.load_labeling_corpus()
    #     valid_x, valid_y = train_x, train_y
    #
    #     model.fit(train_x,
    #               train_y,
    #               x_validate=valid_x,
    #               y_validate=valid_y,
    #               epochs=self.EPOCH_COUNT)
    #
    #     model.evaluate(valid_x, valid_y)
    #     model.evaluate(valid_x, valid_y, truncating=True)
    #     model.predict(valid_x)
    #     model.predict(valid_x, truncating=True)


if __name__ == '__main__':
    unittest.main()



import tensorflow as tf
from tensorflow.python import keras
from tensorflow.python.keras import backend as K

custom_objects = tf.keras.utils.get_custom_objects()

L = keras.layers


class Highway(L.Layer):
    """
      codes from github: https://github.com/batikim09/Keras_highways/blob/master/src/conv2d_highway.py
    """
    activation = None
    transform_gate_bias = None

    def __init__(self, activation='relu', transform_gate_bias=-2, **kwargs):
        self.activation = activation
        self.transform_gate_bias = transform_gate_bias
        super(Highway, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        dim = input_shape[-1]
        self.dense_1 = L.Layer.Dense(units=dim, bias_initializer=L.Layer.Constant(self.transform_gate_bias))
        self.dense_1.build(input_shape)
        self.dense_2 = L.Layer.Dense(units=dim)
        self.dense_2.build(input_shape)
        self.trainable_weights = self.dense_1.trainable_weights + self.dense_2.trainable_weights
        super(Highway, self).build(input_shape)  # Be sure to call this at the end

    def call(self, x):
        dim = K.int_shape(x)[-1]
        transform_gate = self.dense_1(x)
        transform_gate = L.Layer.Activation("sigmoid")(transform_gate)
        carry_gate = L.Layer.Lambda(lambda x: 1.0 - x, output_shape=(dim,))(transform_gate)
        transformed_data = self.dense_2(x)
        transformed_data = L.Layer.Activation(self.activation)(transformed_data)
        transformed_gated = L.Layer.Multiply()([transform_gate, transformed_data])
        identity_gated = L.Layer.Multiply()([carry_gate, x])
        value = L.Layer.Add()([transformed_gated, identity_gated])
        return value

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config['activation'] = self.activation
        config['transform_gate_bias'] = self.transform_gate_bias
        return config



from abc import ABCMeta, abstractmethod

from kolibri.data.corpus.base_stream import BaseStream
from kolibri.metrics.measures import WindowClassificationMeasurements, ClassificationMeasurements, \
    RegressionMeasurements, \
    WindowRegressionMeasurements, MultiTargetClassificationMeasurements, WindowMultiTargetClassificationMeasurements, \
    MultiTargetRegressionMeasurements, \
    WindowMultiTargetRegressionMeasurements
from kolibri.model_trainer import ModelTrainer
from kolibri.utils import constants
from kolibri.visualizations.performance_visualizer import PerformanceVisualizer


class EvaluationDataBuffer(object):
    """ Stores evaluation texts at a given time. It is used to track and distribute
    texts across evaluators and visualizers.
    """

    def __init__(self, data_dict):
        """
        Parameters
        ----------
        data_dict: dict
            Dictionary containing metrics as keys and texts identifiers as values.
        """
        self.data_dict = None
        self.data = {}
        self.sample_id = -1
        self._validate(data_dict)
        self._clear_data()

    def _validate(self, data_dict):
        if isinstance(data_dict, dict):
            if len(data_dict.keys()) > 0:
                self.data_dict = data_dict
            else:
                raise ValueError('data_dict is empty')
        else:
            raise TypeError('data_dict must be a dictionary, received: {}'.format(type(data_dict)))

    def _clear_data(self):
        for metric_id, data_ids in self.data_dict.items():
            self.data[metric_id] = {}
            for data_id in data_ids:
                self.data[metric_id][data_id] = None

    def get_data(self, metric_id, data_id):
        return self.data[metric_id][data_id]

    def update_data(self, sample_id, metric_id, data_id, value):
        if self.sample_id != sample_id:
            # New sample id, clear values
            self._clear_data()
            self.sample_id = sample_id
        self.data[metric_id][data_id] = value


class BaseEvaluator(metaclass=ABCMeta):
    """ The abstract class that works as a base model for all of this framework's
    evaluators. It creates a basic interface that evaluation modules should
    follow in order to use them with all the tools available in scikit-workflow.
    This class should not me instantiated, as none of its methods, except the
    get_class_type, are implemented.
    Raises
    ------
    NotImplementedError: This is an abstract class.
    """

    def __init__(self):
        # Evaluator configuration
        self.n_wait = 0
        self.max_samples = 0
        self.batch_size = 0
        self.pretrain_size = 0
        self.max_time = 0
        self.metrics = []
        self.output_file = None
        self.output_plot = False
        self.is_stream = True
        self.test_size = 0
        self.dynamic_test_set = False
        self.data_points_for_classification = False

        # Metrics
        self.mean_eval_measurements = None
        self.current_eval_measurements = None
        self._data_dict = None
        self._data_buffer = None

        # Misc
        self._method = None
        self._task_type = None
        self._output_type = None
        self._valid_configuration = False
        self.model_names = None
        self.model = None
        self.n_models = 0
        self.stream = None

        self.n_sliding = 0
        self.global_sample_count = 0

    def get_class_type(self):
        return 'evaluator'

    @abstractmethod
    def set_params(self, parameter_dict):
        """ Update parameter names and values via a dictionary.
        Parameters
        ----------
        parameter_dict: dictionary
            A dictionary where the keys are parameters' names and the values
            are the new values for those parameters.
        """
        raise NotImplementedError

    @abstractmethod
    def get_info(self):
        """Collects information about the evaluator.
            Returns
            -------
            string
                Evaluator description.
        """
        raise NotImplementedError

    def _init_evaluation(self, stream, model, model_names=None):
        # First, verify if this is a single evaluation or a comparison between learners.
        if isinstance(model, list):
            self.n_models = len(model)
            for m in model:
                if not hasattr(m, 'predict'):
                    raise NotImplementedError('{} does not have a predict() method.'.format(m))
        else:
            self.n_models = 1
            if isinstance(model, ModelTrainer):
                model = model.pipeline
            elif not hasattr(model, 'predict'):
                raise NotImplementedError('{} does not have a predict() method.'.format(model))

        self.model = model if isinstance(model, list) else [model]
        if isinstance(stream, BaseStream):
            self.stream = stream
        else:
            raise ValueError('{} is not a valid stream type.'.format(stream))

        if model_names is None:
            self.model_names = ['M{}'.format(i) for i in range(self.n_models)]
        else:
            if isinstance(model_names, list):
                if len(model_names) != self.n_models:
                    raise ValueError("Number of model names does not match the number of models.")
                else:
                    self.model_names = model_names
            else:
                raise ValueError("model_names must be a list.")

    def _check_configuration(self):
        # Check stream to infer task type
        if isinstance(self.stream, BaseStream):
            if self.stream.n_targets == 1:
                self._output_type = constants.SINGLE_OUTPUT
            elif self.stream.n_targets > 1:
                self._output_type = constants.MULTI_OUTPUT
            else:
                raise ValueError('Unexpected number of outputs in stream: {}.'.format(self.stream.n_targets))
        else:
            raise ValueError('{} is not a valid stream type.'.format(self.stream))

        # Metrics configuration
        self.metrics = [x.lower() for x in self.metrics]

        for plot in self.metrics:
            if plot not in constants.PLOT_TYPES:
                raise ValueError('Plot type not supported: {}.'.format(plot))

        # Check consistency between output type and metrics and between metrics
        if self._output_type == constants.SINGLE_OUTPUT:
            classification_metrics = set(constants.CLASSIFICATION_METRICS)
            regression_metrics = set(constants.REGRESSION_METRICS)
            evaluation_metrics = set(self.metrics)

            if evaluation_metrics.intersection(classification_metrics) == \
                    evaluation_metrics.intersection(regression_metrics):
                self._task_type = constants.UNDEFINED
                raise ValueError("You need another metric with {}".format(self.metrics))

            elif evaluation_metrics.union(classification_metrics) == classification_metrics or \
                    self.data_points_for_classification:
                self._task_type = constants.CLASSIFICATION
            elif evaluation_metrics.union(regression_metrics) == regression_metrics:
                self._task_type = constants.REGRESSION
            else:
                raise ValueError("Inconsistent metrics {} for {} stream.".format(self.metrics, self._output_type))
        else:
            multi_label_classification_metrics = set(constants.MULTI_TARGET_CLASSIFICATION_METRICS)
            multi_target_regression_metrics = set(constants.MULTI_TARGET_REGRESSION_METRICS)
            evaluation_metrics = set(self.metrics)

            if evaluation_metrics.union(multi_label_classification_metrics) == multi_label_classification_metrics:
                self._task_type = constants.MULTI_TARGET_CLASSIFICATION
            elif evaluation_metrics.union(multi_target_regression_metrics) == multi_target_regression_metrics:
                self._task_type = constants.MULTI_TARGET_REGRESSION
            else:
                raise ValueError("Inconsistent metrics {} for {} stream.".format(self.metrics, self._output_type))

        self._valid_configuration = True

        return self._valid_configuration

    def _check_progress(self, logging, n_samples):
        progress = self.global_sample_count - self.batch_size

        # Update progress
        try:
            if (progress % (n_samples // 20)) == 0:
                logging.info('{}%'.format(progress // (n_samples / 20) * 5))
        except ZeroDivisionError:
            raise ZeroDivisionError("The stream is too small to evaluate. The minimum size is 20 samples.")

    def _init_metrics(self):
        """ Starts up the metrics and statistics watchers. One watcher is created
        for each of the learners to be evaluated.

        """
        self.mean_eval_measurements = []
        self.current_eval_measurements = []

        if self._task_type == constants.CLASSIFICATION:
            for i in range(self.n_models):
                self.mean_eval_measurements.append(ClassificationMeasurements())
                self.current_eval_measurements.append(WindowClassificationMeasurements(window_size=self.n_sliding))

        elif self._task_type == constants.MULTI_TARGET_CLASSIFICATION:
            for i in range(self.n_models):
                self.mean_eval_measurements.append(MultiTargetClassificationMeasurements())
                self.current_eval_measurements.append(WindowMultiTargetClassificationMeasurements(
                    window_size=self.n_sliding))

        elif self._task_type == constants.REGRESSION:
            for i in range(self.n_models):
                self.mean_eval_measurements.append(RegressionMeasurements())
                self.current_eval_measurements.append(WindowRegressionMeasurements(window_size=self.n_sliding))
        elif self._task_type == constants.MULTI_TARGET_REGRESSION:
            for i in range(self.n_models):
                self.mean_eval_measurements.append(MultiTargetRegressionMeasurements())
                self.current_eval_measurements.append(WindowMultiTargetRegressionMeasurements(
                    window_size=self.n_sliding))

        # Evaluation texts buffer
        self._data_dict = {}
        for metric in self.metrics:
            data_ids = [constants.MEAN, constants.CURRENT]
            if metric == constants.TRUE_VS_PREDICTED:
                data_ids = [constants.Y_TRUE, constants.Y_PRED]
            elif metric == constants.DATA_POINTS:
                data_ids = ['X', 'target_values', 'prediction']
            self._data_dict[metric] = data_ids

        self._data_buffer = EvaluationDataBuffer(data_dict=self._data_dict)

    def _update_metrics(self):
        """ Updates the metrics of interest. This function updates the evaluation texts buffer
        which is used to track performance during evaluation.

        The content of the buffer depends on the evaluation task type and metrics selected.

        If more than one model/learner is evaluated at once, texts is stored as lists inside
        the buffer.
        """
        shift = 0
        if self._method == 'prequential':
            shift = -self.batch_size  # Adjust index due to training after testing
        sample_id = self.global_sample_count + shift

        for metric in self.metrics:
            values = [[], []]
            if metric == constants.ACCURACY:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_accuracy())
                    values[1].append(self.current_eval_measurements[i].get_accuracy())

            elif metric == constants.HAMMING_SCORE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_hamming_score())
                    values[1].append(self.current_eval_measurements[i].get_hamming_score())

            elif metric == constants.HAMMING_LOSS:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_hamming_loss())
                    values[1].append(self.current_eval_measurements[i].get_hamming_loss())

            elif metric == constants.EXACT_MATCH:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_exact_match())
                    values[1].append(self.current_eval_measurements[i].get_exact_match())

            elif metric == constants.J_INDEX:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_j_index())
                    values[1].append(self.current_eval_measurements[i].get_j_index())

            elif metric == constants.MSE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_mean_square_error())
                    values[1].append(self.current_eval_measurements[i].get_mean_square_error())

            elif metric == constants.MAE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_average_error())
                    values[1].append(self.current_eval_measurements[i].get_average_error())

            elif metric == constants.AMSE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_average_mean_square_error())
                    values[1].append(self.current_eval_measurements[i].get_average_mean_square_error())

            elif metric == constants.AMAE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_average_absolute_error())
                    values[1].append(self.current_eval_measurements[i].get_average_absolute_error())

            elif metric == constants.ARMSE:
                for i in range(self.n_models):
                    values[0].append(self.mean_eval_measurements[i].get_average_root_mean_square_error())
                    values[1].append(self.current_eval_measurements[i].get_average_root_mean_square_error())

            elif metric == constants.TRUE_VS_PREDICTED:
                y_true = -1
                y_pred = []
                for i in range(self.n_models):
                    t, p = self.mean_eval_measurements[i].get_last()
                    y_true = t  # We only need to keep one true value
                    y_pred.append(p)
                values[0] = y_true
                for i in range(self.n_models):
                    values[1].append(y_pred[i])

            elif metric == constants.DATA_POINTS:
                target_values = self.stream.target_values
                features = {}  # Dictionary containing feature values, using index as key

                y_pred, p = self.mean_eval_measurements[0].get_last()  # Only track one model (first) by default

                X, _ = self.stream.last_sample()
                idx_1 = 0  # TODO let the user choose the feature indices of interest
                idx_2 = 1
                features[idx_1] = X[0][idx_1]
                features[idx_2] = X[0][idx_2]

                values = [None, None, None]
                values[0] = features
                values[1] = target_values
                values[2] = y_pred

            else:
                raise ValueError('Unknown metric {}'.format(metric))

            # Update buffer
            if metric == constants.TRUE_VS_PREDICTED:
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id=constants.Y_TRUE,
                                              value=values[0])
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id=constants.Y_PRED,
                                              value=values[1])
            elif metric == constants.DATA_POINTS:
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id='X',
                                              value=values[0])
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id='target_values',
                                              value=values[1])
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id='predictions',
                                              value=values[2])
            else:
                # Default case, 'mean' and 'current' performance
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id=constants.MEAN,
                                              value=values[0])
                self._data_buffer.update_data(sample_id=sample_id, metric_id=metric, data_id=constants.CURRENT,
                                              value=values[1])

        shift = 0
        if self._method == 'prequential':
            shift = -self.batch_size  # Adjust index due to training after testing
        self._update_outputs(self.global_sample_count + shift)

    def _update_outputs(self, sample_id):
        """ Update outputs of the evaluation. """
        self._update_file()
        if self.show_plot:
            if self.visualizer is not None:
                self.visualizer.on_new_train_step(sample_id, self._data_buffer)

    def _init_file(self):

        if self.output_file is not None:
            with open(self.output_file, 'w+') as f:
                f.write("# TEST CONFIGURATION BEGIN")
                if hasattr(self.stream, 'get_info'):
                    f.write("\n# {}".format(self.stream.get_info()))
                for i in range(self.n_models):
                    if hasattr(self.model[i], 'get_info'):
                        f.write("\n# [{}] {}".format(self.model_names[i], self.model[i].get_info()))
                f.write("\n# {}".format(self.get_info()))
                f.write("\n# TEST CONFIGURATION END")
                header = '\nid'
                for metric in self.metrics:
                    if metric == constants.ACCURACY:
                        for i in range(self.n_models):
                            header += ',mean_acc_[{0}],current_acc_[{0}]'.format(self.model_names[i])
                    elif metric == constants.MSE:
                        for i in range(self.n_models):
                            header += ',mean_mse_[{0}],current_mse_[{0}]'.format(self.model_names[i])
                    elif metric == constants.MAE:
                        for i in range(self.n_models):
                            header += ',mean_mae_[{0}],current_mae_[{0}]'.format(self.model_names[i])
                    elif metric == constants.AMSE:
                        for i in range(self.n_models):
                            header += ',mean_amse_[{0}],current_amse_[{0}]'.format(self.model_names[i])
                    elif metric == constants.AMAE:
                        for i in range(self.n_models):
                            header += ',mean_amae_[{0}],current_amae_[{0}]'.format(self.model_names[i])
                    elif metric == constants.ARMSE:
                        for i in range(self.n_models):
                            header += ',mean_armse_[{0}],current_armse_[{0}]'.format(self.model_names[i])
                    elif metric == constants.TRUE_VS_PREDICTED:
                        header += ',true_value'
                        for i in range(self.n_models):
                            header += ',predicted_value_[{0}]'.format(self.model_names[i])
                    else:
                        for i in range(self.n_models):
                            header += ',mean_{0}_[{1}],current_{0}_[{1}]'.format(metric, self.model_names[i])
                f.write(header)

    def _update_file(self):
        if self.output_file is not None:
            # Note: Must follow order set in _init_file()
            line = str(self._data_buffer.sample_id)
            for metric in self.metrics:
                if metric == constants.TRUE_VS_PREDICTED:
                    true_value = self._data_buffer.get_data(metric_id=metric, data_id=constants.Y_TRUE)
                    pred_values = self._data_buffer.get_data(metric_id=metric, data_id=constants.Y_PRED)
                    line += ',{:.6f}'.format(true_value)
                    for i in range(self.n_models):
                        line += ',{:.6f}'.format(pred_values[i])
                else:
                    mean_values = self._data_buffer.get_data(metric_id=metric, data_id=constants.MEAN)
                    current_values = self._data_buffer.get_data(metric_id=metric, data_id=constants.CURRENT)
                    values = (mean_values, current_values)
                    for i in range(self.n_models):
                        line += ',{:.6f},{:.6f}'.format(values[0][i], values[1][i])

            with open(self.output_file, 'a') as f:
                f.write('\n' + line)

    def _reset_globals(self):
        self.global_sample_count = 0

    def evaluation_summary(self, logging, start_time, end_time):
        if end_time - start_time > self.max_time:
            logging.info('Time limit reached. Evaluation stopped.')
            logging.info('Evaluation time:     {:.2f} s'.format(self.max_time))
        else:
            logging.info('Evaluation time:     {:.2f} s'.format(end_time - start_time))
        logging.info('Processed samples: {}'.format(self.global_sample_count))
        logging.info('Mean performance:')
        for i in range(self.n_models):
            if constants.ACCURACY in self.metrics:
                logging.info('{} - Accuracy     : {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.ACCURACY, data_id=constants.MEAN)[i]))
            if constants.HAMMING_SCORE in self.metrics:
                logging.info('{} - Hamming score: {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.HAMMING_SCORE, data_id=constants.MEAN)[i]))
            if constants.HAMMING_LOSS in self.metrics:
                logging.info('{} - Hamming loss : {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.HAMMING_LOSS, data_id=constants.MEAN)[i]))
            if constants.EXACT_MATCH in self.metrics:
                logging.info('{} - Exact matches: {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.EXACT_MATCH, data_id=constants.MEAN)[i]))
            if constants.J_INDEX in self.metrics:
                logging.info('{} - Jaccard index: {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.J_INDEX, data_id=constants.MEAN)[i]))
            if constants.MSE in self.metrics:
                logging.info('{} - MSE          : {:.4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.MSE, data_id=constants.MEAN)[i]))
            if constants.MAE in self.metrics:
                logging.info('{} - MAE          : {:4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.MAE, data_id=constants.MEAN)[i]))
            if constants.AMSE in self.metrics:
                logging.info('{} - AMSE          : {:4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.AMSE, data_id=constants.MEAN)[i]))
            if constants.AMAE in self.metrics:
                logging.info('{} - AMAE          : {:4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.AMAE, data_id=constants.MEAN)[i]))
            if constants.ARMSE in self.metrics:
                logging.info('{} - ARMSE          : {:4f}'.format(
                    self.model_names[i],
                    self._data_buffer.get_data(metric_id=constants.ARMSE, data_id=constants.MEAN)[i]))

    def get_measurements(self, model_idx=None):
        """ Get measurements from the evaluation.
        Parameters
        ----------
        model_idx: int, optional (Default=None)
            Indicates the index of the model as defined in `evaluate(model)`.
            If None, returns a list with the measurements for each model.
        Returns
        -------
        tuple (mean, current)
        Mean and Current measurements. If model_idx is None, each member of the tuple
         is a a list with the measurements for each model.
        Raises
        ------
        IndexError: If the index is invalid.
        """
        if model_idx is None:
            return self.mean_eval_measurements, self.current_eval_measurements
        else:
            try:
                # Check index
                _ = self.mean_eval_measurements[model_idx]
                _ = self.current_eval_measurements[model_idx]
            except IndexError:
                print('Model index {} is invalid'.format(model_idx))
                return None, None
            return self.mean_eval_measurements[model_idx], self.current_eval_measurements[model_idx]

    def get_mean_measurements(self, model_idx=None):
        """ Get mean measurements from the evaluation.
        Parameters
        ----------
        model_idx: int, optional (Default=None)
            Indicates the index of the model as defined in `evaluate(model)`.
            If None, returns a list with the measurements for each model.
        Returns
        -------
        measurements or list
        Mean measurements. If model_idx is None, returns a list with the measurements
         for each model.
        Raises
        ------
        IndexError: If the index is invalid.
        """
        measurements, _ = self.get_measurements(model_idx)
        return measurements

    def get_current_measurements(self, model_idx=None):
        """ Get current measurements from the evaluation (measured on last `n_wait` samples).
        Parameters
        ----------
        model_idx: int, optional (Default=None)
            Indicates the index of the model as defined in `evaluate(model)`.
            If None, returns a list with the measurements for each model.
        Returns
        -------
        measurements or list
        Current measurements. If model_idx is None, returns a list with the measurements
         for each model.
        Raises
        ------
        IndexError: If the index is invalid.
        """
        _, measurements = self.get_measurements(model_idx)
        return measurements

    def _init_plot(self):
        """ Initialize plot to display the evaluation results.

        """
        if self.show_plot:
            self.visualizer = PerformanceVisualizer(task_type=self._task_type,
                                                    n_wait=self.n_sliding,
                                                    dataset_name=self.stream.get_data_info(),
                                                    metrics=self.metrics,
                                                    n_models=self.n_models,
                                                    model_names=self.model_names,
                                                    data_dict=self._data_dict)



import logging

from kolibri.application.datastore.index_tagger import PosLemmaIndexTagger

logger = logging.getLogger(__name__)


class DatabaseAdapter(object):
    """
    Abstract class that represents the interface
    that all database adapters have to implement.
    """

    def __init__(self, language='en'):
        """
        :param str language: The language for the tagger .
        """

        self.tagger = PosLemmaIndexTagger(language)

    def get_model(self, model_name):
        """
        Return the model class for a given model name.

        model_name is case insensitive.
        """
        get_model_method = getattr(self, 'get_%s_model' % (
            model_name.lower(),
        ))

        return get_model_method()

    def get_object(self, object_name):
        """
        Return the class for a given object name.

        object_name is case insensitive.
        """
        get_model_method = getattr(self, 'get_%s_object' % (
            object_name.lower(),
        ))

        return get_model_method()

    def get_document_object(self):
        from kolibri.document import Document

        DocumentModel = self.get_model('document')

        Document.document_field_names.extend(
            DocumentModel.extra_document_field_names
        )

        return Document

    def count(self):
        """
        Return the number of entries in the database.
        """
        raise NotImplementedError(
            'The `count` method is not implemented by this adapter.'
        )

    def remove(self, document_text):
        """
        Removes the document that matches the input text.
        """
        raise NotImplementedError(
            'The `remove` method is not implemented by this adapter.'
        )

    def filter(self, **kwargs):
        """
        Returns a list of objects from the database.
        The kwargs parameter can contain any number
        of attributes.

        :param page_size: The maximum number of records to load into
            memory at once when returning results.
            Defaults to 1000

        :param order_by: The field that will be sorted

        :param tags: A list of tags. When specified, the results will only
            include documents that have a tag in the provided list.

        :param exclude_text: documents that contain this text will not be retrieved.
            Defaults to None

        :param exclude_text_words: document that contain any word form the list will
        not be retrieved


        :param search_text_contains: If the ``search_text`` field of a
            document from the provided stringù; it will be returned.
        """
        raise NotImplementedError(
            'The `filter` method is not implemented by this adapter.'
        )

    def create(self, **kwargs):
        """
        Creates a new document matching the keyword arguments specified.
        Returns the created document.
        """
        raise NotImplementedError(
            'The `create` method is not implemented by this adapter.'
        )

    def create_multiple(self, documents):
        """
        Creates multiple document entries.
        """
        raise NotImplementedError(
            'The `create_many` method is not implemented by this adapter.'
        )

    def update(self, document):
        """
        Modifies an entry in the database.
        Creates an entry if one does not exist.
        """
        raise NotImplementedError(
            'The `update` method is not implemented by this adapter.'
        )

    def get_random(self):
        """
        Returns a random document from the database.
        """
        raise NotImplementedError(
            'The `get_random` method is not implemented by this adapter.'
        )

    def drop(self):
        """
        Drop the database attached to a given adapter.
        """
        raise NotImplementedError(
            'The `drop` method is not implemented by this adapter.'
        )



import logging

from kolibri import settings
from kolibri.config import TaskType
from kolibri.dnn.embeddings import DefaultEmbedding, WordEmbedding
from kolibri.dnn.tasks.dnn_estimator import DnnEstimator
from kolibri.dnn.tasks.text.classification.models import get_model

logger = logging.getLogger(__name__)

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"
DNN_MODEL_FILE_NAME = "classifier_dnn"


class DnnTextClassEstimator(DnnEstimator):
    """classifier using the sklearn framework"""

    name = 'dnn_classifier'

    provides = ["classification", "target_ranking"]

    requires = ["text_features"]

    defaults = {

        # the models used in the classifier if several models are given they will be combined
        "models": "bilstm",
        "embeddings": "default",
        "multi-label": False,
        "sequence_length": 'auto',
        "epochs": 1,
        "loss": 'categorical_crossentropy',
        "class-weight": False,

    }

    def __init__(self, component_config=None):

        """Construct a new class classifier using the sklearn framework."""
        super().__init__(component_config)

        if self.component_config['embeddings'] == 'default':
            self.embeddings = DefaultEmbedding(task=TaskType.CLASSIFICATION,
                                               sequence_length=self.component_config["sequence_length"],
                                               indexer=self.indexer)
        elif self.component_config['embeddings'] == 'word':
            self.embeddings = WordEmbedding(w2v_path=self.component_config["embedding_path"],
                                            task=TaskType.CLASSIFICATION,
                                            sequence_length=self.component_config["sequence_length"],
                                            indexer=self.indexer)
        elif self.component_config['embeddings'] == None:
            self.embeddings = None

        self.clf = get_model(self.component_config['models'], embedding=self.embeddings,
                             hyper_parameters=self.component_config)
        self.classifier_type = type(self.clf)

    @classmethod
    def required_packages(cls):
        return ["tensorflow"]

    def process(self, document, **kwargs):
        """Return the most likely class and its probability for a document."""
        raw_results = None
        if not self.clf:
            # component is either not trained or didn't
            # receive enough training texts
            target = None
            target_ranking = []
        else:
            X = document.tokens
            raw_results = self.clf.predict([X], top_k=settings.modeling['TARGET_RANKING_LENGTH'])

            if len(raw_results) > 0:

                target = {"name": raw_results[0]['label'], "confidence": raw_results[0]['confidence']}

                target_ranking = raw_results[0]['confidence']
            else:
                target = {"name": None, "confidence": 0.0}
                target_ranking = []

        document.label = target
        document.raw_prediction_results = raw_results
        document.set_output_property("raw_prediction_results")
        document.set_output_property("label")
        document.target_ranking = target_ranking
        document.set_output_property("target_ranking")



import random
import unittest

from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus
from kolibri.dnn.utils import load_data_object
from kolibri.indexers import TextIndexer, LabelIndexer
from kolibri.tokenizer import WordTokenizer

tokenizer = WordTokenizer()


class TestTextIndexer(unittest.TestCase):
    def test_text_indexer(self):
        corpus = ConsumerComplaintsCorpus()
        x_set, y_set = corpus.get_data(nb_samples=500)
        x_set = tokenizer.transform(x_set)
        x_samples = random.sample(x_set, 5)
        text_processor = TextIndexer(min_count=1)
        text_processor.build_vocab(x_set, y_set)
        text_idx = text_processor.transform(x_samples)

        text_info_dict = text_processor.to_dict()
        text_processor2: TextIndexer = load_data_object(text_info_dict)

        text_idx2 = text_processor2.transform(x_samples)
        sample_lengths = [len(i) for i in x_samples]

        assert (text_idx2 == text_idx).all()
        assert text_processor.inverse_transform(text_idx, lengths=sample_lengths) == x_samples
        assert text_processor2.inverse_transform(text_idx2, lengths=sample_lengths) == x_samples

    def test_label_indexer(self):
        corpus = ConsumerComplaintsCorpus()
        x_set, y_set = corpus.get_data(nb_samples=500)
        x_set = tokenizer.transform(x_set)
        text_processor = LabelIndexer(build_vocab_from_labels=True, min_count=1)
        text_processor.build_vocab(x_set, y_set)

        samples = random.sample(y_set, 20)

        text_idx = text_processor.transform(samples, )

        text_info_dict = text_processor.to_dict()

        text_processor2: LabelIndexer = load_data_object(text_info_dict)

        text_idx2 = text_processor2.transform(samples, )
        lengths = [len(i) for i in samples]
        assert (text_idx2 == text_idx).all()
        assert text_processor2.inverse_transform(text_idx, lengths=lengths) == samples
        assert text_processor2.inverse_transform(text_idx2, lengths=lengths) == samples


if __name__ == "__main__":
    pass



from copy import copy

import numpy as np
from scipy.spatial.distance import hamming
from sympy.combinatorics.graycode import GrayCode

from kolibri import settings
from kolibri.tasks.classification.estimator import Estimator

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"


class EcocEstimator(Estimator):
    """
    ECOC meta classifier
    ref: http://www.cs.cmu.edu/~aberger/pdf/ecoc.pdf
    """
    name = 'ecoc_classifier'

    @staticmethod
    def softmax(x):
        """Compute softmax values for each sets of scores in x."""
        e_x = np.exp(x - np.max(x))
        return e_x / e_x.sum(axis=0)  # only difference

    def __init__(self, config={}, base_estimator=None, indexer=None, code_size=1.0, random_state=42):

        super().__init__(component_config=config, classifier=base_estimator, indexer=indexer)
        self.code_size = code_size
        self.error_code = None
        self.estimators = None

        self.random_state = random_state

    def _generate_gray_codes(self, code_size):
        gray = np.array([[int(j) for j in i] for i in GrayCode(code_size).generate_gray()])
        total_values_sum = gray.sum(axis=1)
        to_delete = [total_values_sum.argmax(), total_values_sum.argmin()]
        return np.delete(gray, to_delete, axis=0)

    def _hamming_corrmat(self, gray):
        result = []

        # TODO: reimplement a method, very slow and unoptimized
        for i in gray:
            row = []
            for j in gray:
                row.append(hamming(i, j))
            result.append(row)

        return np.array(result)

    def _greedy_search(self, a, size=None):
        if size is None:
            size = int(np.log2(a.shape[0] + 1))

        result = []
        current = np.random.randint(0, a.shape[0])
        result.append(current)

        for i in range(size - 1):
            row_sum = a[:, result].sum(1)
            min_border = np.percentile(row_sum, 85)
            valid_vals = row_sum[row_sum >= min_border]
            max_index = np.random.choice(
                np.argwhere(np.isin(row_sum, valid_vals)).flatten()
                , 1).squeeze()

            result.append(int(max_index))
            a[current, max_index] = -1
            a[max_index, current] = -1
            current = max_index
        return result

    def fit(self, X, y, X_val=None, y_val=None):
        if y is None:
            raise Exception('y should be initialized')

        unique_classes = np.unique(y)

        self.indexer.build_vocab(None, y)
        _y = np.array(self.indexer.transform(y, ))
        #        if y_val is not None:
        #            y_val=np.array(self.indexer.numerize_label_sequences(y_val))

        classes_size = unique_classes.shape[0]
        code_size = int(classes_size * self.code_size)

        if code_size < np.log2(classes_size):
            print('inapropriate code size, try to adjust it')

        if code_size > 8:
            # if code size is too small random initialization
            # is not robust enough, it is more rational to use
            # gray codes to generate codes with maximum distance

            np.random.seed(self.random_state)
            self.error_code = np.random.randint(2, size=(classes_size, code_size))
            while set([0, classes_size]).intersection(self.error_code.sum(0)) \
                    or set([0, code_size]).intersection(self.error_code.sum(1)):
                self.error_code = np.random.randint(2, size=(classes_size, code_size))
        else:
            gray_codes = self._generate_gray_codes(classes_size)
            gray_corrmat = self._hamming_corrmat(gray_codes)
            codes_id = self._greedy_search(gray_corrmat, code_size)
            self.error_code = np.array(gray_codes[codes_id]).T

        # iterate over error codez
        estimators = []
        for i in range(self.error_code.shape[1]):
            print('fitting %d-th classifier out of %d' % (i + 1, code_size))

            classifier_mapping = self.error_code[:, i]
            # print(classifier_mapping)

            positive = np.argwhere(classifier_mapping == 1).flatten()
            negative = np.argwhere(classifier_mapping == 0).flatten()

            _y_i = _y.copy()

            _y_i[np.isin(_y_i, positive)] = -2
            _y_i[np.isin(_y_i, negative)] = -1
            _y_i[_y_i == -2] = 1

            estimator = copy(self.clf)

            estimator.fit(X, _y_i)

            estimators.append(estimator)

        self.estimators = estimators

        self.evaluate(X_val, y_val)

    def predict(self, X):
        """
        Extended ECOC model intefrace, implementation is based on
        http://www.cs.cmu.edu/~aberger/pdf/ecoc.pdf
        :param X: classification input (can be a sparse matrix)
        :return: numpy array of classification results
        """

        if not hasattr(X, 'shape'):
            raise Exception('X must be a numpy object')

        predicted = []
        for i in range(len(self.estimators)):
            predicted.append(self.estimators[i].predict(X) * -1)

        predicted = np.array(predicted).T

        scores = []
        for i in range(predicted.shape[0]):
            tiled_predictions = np.tile(predicted[i], (self.error_code.shape[0], 1))

            class_score = np.sum(np.abs(self.error_code - tiled_predictions), axis=1)
            class_score = np.argmax(class_score)

            scores.append(class_score)

        return self.indexer.inverse_transform(scores)

    def predict_proba(self, X, y=None):
        """
        Extended ECOC model intefrace, implementation is based on
        http://www.cs.cmu.edu/~aberger/pdf/ecoc.pdf
        :param X: classification input (can be a sparse matrix)
        :return: numpy array of classification results
        """
        if not hasattr(X, 'shape'):
            raise Exception('X must be a numpy object')

        predicted = []
        for i in range(len(self.estimators)):
            predicted.append(self.estimators[i].predict(X) * -1)

        predicted = np.array(predicted).T

        scores = []
        for i in range(predicted.shape[0]):
            tiled_predictions = np.tile(predicted[i], (self.error_code.shape[0], 1))

            class_score = np.sum(np.abs(self.error_code - tiled_predictions), axis=1)
            class_score = EcocEstimator.softmax(class_score)

            scores.append(class_score)
            scores = np.array(scores)
        sorted_indices = np.fliplr(np.argsort(scores, axis=1))

        return sorted_indices, scores[:, sorted_indices]

    def predict_top_n(self, X, n=5):
        """
        Predict top N most relevant classes for each record
        :param X: classification input (can be a sparse matrix)
        :param n: how much top results should be given
        :return: numpy matrix of top predictions
        """
        result = self.predict_proba(X)
        sorted_probas = copy(result)
        sorted_probas.sort(axis=1)
        sorted_probas = sorted_probas[:, -n:][:, ::-1]
        class_mapper = np.vectorize(lambda x: self.id2class[x])

        classes = class_mapper(result.argsort(axis=1)[:, -n:][:, ::-1])

        sorted_probas = [[float(j) for j in i] for i in sorted_probas.tolist()]
        classes = [[j for j in i] for i in classes.tolist()]

        return list(map(lambda x: dict(zip(x[0], x[1])), zip(classes, sorted_probas)))

    def shrink_prediction(self, X, y, n=5):
        """
        Reduce prediction of predict_top_n with the following rule:
        if the true result is in top, then replace the prediction with true result
        otherwise replace prediction with the most significant prediction
        :param X: classification input (can be a sparse matrix)
        :param y: real y_values
        :param n: how much top results should be given
        :return: numpy array of shrinked predictions
        """
        top_args = self.predict_top_n(X, n=n)
        shrinked = []
        for i, r in zip(top_args, y):
            if r in i:
                shrinked.append(r)
            else:
                keys, vals = list(zip(*i.items()))
                shrinked.append(keys[np.argmax(vals)])

        return shrinked

    def process(self, document, **kwargs):
        """Return the most likely class and its probability for a document."""

        if not self.clf:
            # component is either not trained or didn't
            # receive enough training texts
            target = None
            target_ranking = []
        else:
            X = document.vector
            class_ids, probabilities = self.predict_proba(X)
            classes = self.indexer.inverse_transform(np.ravel(class_ids))
            # `predict` returns a matrix as it is supposed
            # to work for multiple examples as well, hence we need to flatten
            classes, probabilities = classes, probabilities.flatten()

            if len(classes) > 0 and probabilities.size > 0:
                ranking = list(zip(list(classes),
                                   list(probabilities)))[:settings.modeling['TARGET_RANKING_LENGTH']]

                target = {"name": classes[0], "confidence": probabilities[0]}

                target_ranking = [{"name": class_name, "confidence": score}
                                  for class_name, score in ranking]
            else:
                target = {"name": None, "confidence": 0.0}
                target_ranking = []
        #            self.clf.classes_ = self.indexer.la

        document.label = target
        document.set_output_property("raw_prediction_results")
        document.set_output_property("label")
        document.target_ranking = target_ranking
        document.set_output_property("target_ranking")

    # @classmethod
    # def load(cls, model_dir=None, model_metadata=None, cached_component=None, **kwargs):
    #
    #     meta = model_metadata.for_component(cls.name)
    #     file_name = meta.get("classifier_file", KOLIBRI_MODEL_FILE_NAME)
    #     classifier_file = os.path.join(model_dir, file_name)
    #
    #     if os.path.exists(classifier_file):
    #         model = joblib.load(classifier_file)
    #
    #         nb_classifiers=model.error_code.shape[1]
    #         for i in range(nb_classifiers):
    #             estimator_file = os.path.join(model_dir, "classfier_" + str(i) + ".pkl")
    #             estimator = joblib.load(estimator_file)
    #             model.estimators.append(estimator)
    #         return model
    #     else:
    #         return cls(meta)
    #
    # def persist(self, model_dir):
    #     """Persist this model into the passed directory."""
    #
    #     meta_tclassifier_file = os.path.join(model_dir, KOLIBRI_MODEL_FILE_NAME)
    #     joblib.dump(self, meta_tclassifier_file)
    #
    #     for i, estimator in enumerate(self.estimators):
    #         classifier_file=os.path.join(model_dir, "classfier_"+str(i)+".pkl")
    #         joblib.dump(estimator, classifier_file)
    #
    #
    #     return {
    #         "classifier_file": KOLIBRI_MODEL_FILE_NAME,
    #         "accuracy": self.accuracy,
    #         "recall": self.recall,
    #         "f1": self.f1
    #             }



import logging

from kolibri.tasks.classification.estimator import Estimator

logger = logging.getLogger(__name__)

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"


class SklearnEstimator(Estimator):
    """classifier using the sklearn framework"""

    name = 'sklearn_classifier'

    def __init__(self, component_config=None, classifier=None, indexer=None):
        """Construct a new class classifier using the sklearn framework."""
        super().__init__(component_config=component_config, classifier=classifier, indexer=indexer)

    def fit(self, X, y, X_val=None, y_val=None):

        self.indexer.build_vocab(None, y)
        y = self.indexer.transform(y, )
        #        if y_val is not None:
        #            y_val = self.indexer.transform(y_val)

        if self.component_config['priors-thresolding']:
            self.compute_priors(y)

        if self.sampler:

            Xt, yt = self.sampler.fit_resample(X, y)

            self.clf.fit(Xt, yt)
        else:
            self.clf.fit(X, y)

        if X_val is not None and y_val is not None:
            self.evaluate(X_val, y_val)



import os
import random
import tempfile
import time
import unittest

from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus
from kolibri.dnn.embeddings import DefaultEmbedding
from kolibri.dnn.tasks.text.classification import BiGRU_Model
from kolibri.dnn.utils import load_data_object
from kolibri.indexers import TextIndexer
from kolibri.logger import logger
from kolibri.tokenizer import WordTokenizer

sample_count = 50


class TestDefaultEmbedding(unittest.TestCase):

    def build_embedding(self):
        embedding = DefaultEmbedding()
        return embedding

    def test_base_cases(self):
        tokenizer = WordTokenizer()
        embedding = self.build_embedding()
        corpus = ConsumerComplaintsCorpus()
        x, y = corpus.get_data(nb_samples=500)
        x = tokenizer.transform(x)
        indexer = TextIndexer()
        indexer.build_vocab(x, y)
        embedding.setup_text_processor(indexer)

        samples = random.sample(x, sample_count)
        res = embedding.embed(samples)
        max_len = max([len(i) for i in samples]) + 2

        if embedding.max_position is not None:
            max_len = embedding.max_position

        assert res.shape == (len(samples), max_len, embedding.embedding_size)

        # Test Save And Load
        embed_dict = embedding.to_dict()
        embedding2 = load_data_object(embed_dict)
        embedding2.setup_text_processor(indexer)
        assert embedding2.embed(samples).shape == (len(samples), max_len, embedding.embedding_size)

    def test_with_model(self):
        tokenizer = WordTokenizer()

        corpus = ConsumerComplaintsCorpus()

        x, y = corpus.get_data(nb_samples=500)
        x = tokenizer.transform(x)
        embedding = self.build_embedding()

        model = BiGRU_Model(embedding=embedding)
        model.build_model(x, y)
        model_summary = []
        embedding.embed_model.summary(print_fn=lambda x: model_summary.append(x))
        logger.debug('\n'.join(model_summary))

        model.fit(x, y, epochs=1)

        model_path = os.path.join(tempfile.gettempdir(), str(time.time()))
        model.save(model_path)


if __name__ == "__main__":
    unittest.main()
